import torch
import torch.nn as nn
import torch.nn.functional as F
from .crf import CRF
from transformers.modeling_bert import BertPreTrainedModel, BertModel
from .linears import PoolerEndLogits, PoolerStartLogits
from torch.nn import CrossEntropyLoss

class BertSoftmaxForNer(BertPreTrainedModel):
    def __init__(self, config):
        super(BertSoftmaxForNer, self).__init__(config)
        self.num_labels = config.num_labels
        self.bert = BertModel(config)
        self.dropout = nn.Dropout(config.hidden_dropout_prob)
        self.classifier = nn.Linear(config.hidden_size, config.num_labels)
        self.init_weights()

    def forward(self, input_ids, attention_mask=None, token_type_ids=None,
                position_ids=None, head_mask=None, labels=None):
        outputs = self.bert(input_ids,
                            attention_mask=attention_mask,
                            token_type_ids=token_type_ids,
                            position_ids=position_ids,
                            head_mask=head_mask)
        sequence_output = outputs[0]
        sequence_output = self.dropout(sequence_output)
        logits = self.classifier(sequence_output)
        outputs = (logits,) + outputs[2:]  # add hidden states and attention if they are here
        if labels is not None:
            loss_fct = CrossEntropyLoss(ignore_index=0)
            # Only keep active parts of the loss
            if attention_mask is not None:
                active_loss = attention_mask.view(-1) == 1
                active_logits = logits.view(-1, self.num_labels)[active_loss]
                active_labels = labels.view(-1)[active_loss]
                loss = loss_fct(active_logits, active_labels)
            else:
                loss = loss_fct(logits.view(-1, self.num_labels), labels.view(-1))
            outputs = (loss,) + outputs
        return outputs  # (loss), scores, (hidden_states), (attentions)

class BertCrfForNer(BertPreTrainedModel):
    def __init__(self, config, label2id, device):
        super(BertCrfForNer, self).__init__(config)
        self.bert = BertModel(config)
        self.dropout = nn.Dropout(config.hidden_dropout_prob)
        self.classifier = nn.Linear(config.hidden_size, len(label2id))
        self.crf = CRF(tagset_size=len(label2id), tag_dictionary=label2id, device=device, is_bert=True)
        self.init_weights()

    def forward(self, input_ids, token_type_ids=None, attention_mask=None,labels=None,input_lens=None):
        outputs = self.bert(input_ids, token_type_ids, attention_mask)
        sequence_output = outputs[0]
        sequence_output = self.dropout(sequence_output)
        logits = self.classifier(sequence_output)
        outputs = (logits,)
        if labels is not None:
            loss = self.crf.calculate_loss(logits, tag_list=labels, lengths=input_lens)
            outputs =(loss,)+outputs
        return outputs # (loss), scores

class BertSpanForNer(BertPreTrainedModel):
    def __init__(self, config,):
        super(BertSpanForNer, self).__init__(config)
        self.soft_label = config.soft_label
        self.num_labels = config.num_labels
        self.bert = BertModel(config)
        self.dropout = nn.Dropout(config.hidden_dropout_prob)
        self.start_fc = PoolerStartLogits(config.hidden_size, self.num_labels)
        if self.soft_label:
            self.end_fc = PoolerEndLogits(config.hidden_size + self.num_labels, self.num_labels)
        else:
            self.end_fc = PoolerEndLogits(config.hidden_size + 1, self.num_labels)
        self.init_weights()

    def forward(self, input_ids, token_type_ids=None, attention_mask=None, start_positions=None,end_positions=None):
        outputs = self.bert(input_ids, token_type_ids, attention_mask)
        sequence_output = outputs[0]
        sequence_output = self.dropout(sequence_output)
        start_logits = self.start_fc(sequence_output)
        if start_positions is not None and self.training:
            if self.soft_label:
                batch_size = input_ids.size(0)
                seq_len = input_ids.size(1)
                label_logits = torch.FloatTensor(batch_size, seq_len, self.num_labels)
                label_logits.zero_()
                label_logits = label_logits.to(input_ids.device)
                label_logits.scatter_(2, start_positions.unsqueeze(2), 1)
            else:
                label_logits = start_positions.unsqueeze(2).float()
        else:
            label_logits = F.softmax(start_logits, -1)
            if not self.soft_label:
                label_logits = torch.argmax(label_logits, -1).unsqueeze(2).float()
        end_logits = self.end_fc(sequence_output, label_logits)
        outputs = (start_logits, end_logits,) + outputs[2:]

        if start_positions is not None and end_positions is not None:
            start_logits = start_logits.view(-1, self.num_labels)
            end_logits = end_logits.view(-1, self.num_labels)
            active_loss = attention_mask.view(-1) == 1
            active_start_logits = start_logits[active_loss]
            active_start_labels = start_positions.view(-1)[active_loss]
            active_end_logits = end_logits[active_loss]
            active_end_labels = end_positions.view(-1)[active_loss]

            loss_fct = CrossEntropyLoss()
            start_loss = loss_fct(active_start_logits, active_start_labels)
            end_loss = loss_fct(active_end_logits, active_end_labels)
            total_loss = (start_loss + end_loss) / 2
            outputs = (total_loss,) + outputs
        return outputs




#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os
from tqdm import tqdm
from multiprocessing import Pool, cpu_count, Process, Manager, Queue

from loguru import logger


def is_master_process(args):
    return args.local_rank in [-1, 0]


def is_leader_process(args):
    return args.local_rank == 0


def is_member_process(args):
    return args.local_rank > 0


def is_single_process(args):
    return args.local_rank == -1


def is_multi_processes(args):
    return args.local_rank >= 0


def barrier_member_processes(args):
    #  if args.local_rank not in [-1, 0]:
    if is_member_process(args):
        torch.distributed.barrier()


def barrier_leader_process(args):
    #  if args.local_rank == 0:
    if is_leader_process(args):
        torch.distributed.barrier()


def myinstance(process_args, a, b, c):
    i, num_processes = process_args
    logger.debug(f"instance {i}/{num_processes}, {a}, {b}, {c}")


def myinstance1(process_args):
    i, num_processes = process_args
    logger.debug(f"instance {i}/{num_processes}")


def myinstance2(process_args, a):
    i, num_processes = process_args
    logger.debug(f"instance {i}/{num_processes}, {a}")


def split_list_for_multiprocessing(items,
                                   max_folds=os.cpu_count(),
                                   min_items=0):
    """
    将列表items划分成几个子列表，用于多进程列表数据拆分。
    max_folds: 最多划分的份数，缺省值0表示取CPU数目。
    min_items: 每份最少条数，缺省值0表示自动划分。
    返回：拆分后子列表起止序号对的列表。
    """

    assert max_folds != 0 or min_items != 0
    indices = []
    total_items = len(items)
    if total_items == 0:
        return []

    num_folds = max_folds
    #  if num_folds == 0:
    #      return []

    if total_items <= min_items:
        indices = [(0, total_items)]
    else:
        if num_folds == 0:
            num_folds = total_items // min_items
        fold_items = total_items // num_folds
        if fold_items == 0:
            indices = [(0, total_items)]
        else:
            if fold_items < min_items:
                num_folds = total_items // min_items
            fold_items = total_items // num_folds
            indices = [(fold_items * i, fold_items * (i + 1))
                       for i in range(num_folds)]
            indices[-1] = (fold_items * (num_folds - 1), total_items)

    return indices


def run_instances(instance_func,
                  input_list,
                  threads=0,
                  chunksize=32,
                  initialize_func=None,
                  initargs=None):
    if threads <= 0:
        threads = cpu_count()
    threads = min(threads, cpu_count())
    #  with Pool(threads, initialize=initialize_func, initargs=initargs) as P:
    with Pool(threads) as P:
        return_values = list(
            tqdm(
                P.imap(instance_func, input_list, chunksize=chunksize),
                total=len(input_list),
            ))
        return return_values


def start_processes(my_function, args, num_processes=0):
    if num_processes == 0:
        num_processes = os.cpu_count()
    child_processes = []
    return_values = []

    manager = Manager()
    shared_dict = manager.dict()
    for i in range(num_processes):
        logger.debug(f"Start process {i}/{num_processes}")
        child_process = Process(target=my_function,
                                args=[[i, num_processes, shared_dict]] +
                                list(args))
        child_process.start()
        child_processes.append(child_process)
    for process_id, child_process in enumerate(child_processes):
        child_process.join()

    for process_id, child_process in enumerate(child_processes):
        #  v = shared_dict[process_id]
        v = shared_dict.get(process_id, [])
        if len(v) == 0:
            logger.warning(f"Process {process_id} return {len(v)} values.")
        else:
            logger.info(f"Process {process_id} return {len(v)} values.")
        return_values.append(v)
    logger.info(
        f"Exit start_processes. num_processes = {num_processes}), num_shared_dicts = {len(shared_dict)}"
    )

    return return_values


def start_processes_by_queue(my_function, args, num_processes=0):
    if num_processes == 0:
        num_processes = os.cpu_count()
    child_processes = []
    return_values = []
    queues = []
    for i in range(num_processes):
        logger.debug(f"Start process {i}/{num_processes}")
        q = Queue()
        queues.append(q)
        child_process = Process(target=my_function,
                                args=[[i, num_processes, q]] + list(args))
        child_process.start()
        child_processes.append(child_process)
    logger.warning(f"end start Processes. num_queues = {len(queues)}")
    for process_id, child_process in enumerate(child_processes):
        q = queues[process_id]
        v = q.get()
        return_values.append(v)
        child_process.join()
    logger.info(
        f"Exit start_processes. num_processes = {num_processes}), num_queues = {len(queues)}"
    )

    return return_values


def test_split_list_for_multiprocessing_for_multiprocessing():
    assert split_list_for_multiprocessing([1, 2, 3, 4, 5, 6, 7],
                                          max_folds=3) == [(0, 2), (2, 4),
                                                           (4, 7)]
    assert split_list_for_multiprocessing([1, 2, 3, 4, 5, 6],
                                          max_folds=3) == [(0, 2), (2, 4),
                                                           (4, 6)]
    assert split_list_for_multiprocessing([1, 2, 3, 4, 5, 6, 7],
                                          min_items=3) == [(0, 7)]
    assert split_list_for_multiprocessing([1, 2, 3, 4, 5, 6],
                                          min_items=3) == [(0, 6)]
    assert split_list_for_multiprocessing([1, 2, 3, 4, 5, 6, 7]) == [(0, 7)]
    #
    assert split_list_for_multiprocessing([1, 2, 3, 4, 5, 6, 7],
                                          max_folds=3,
                                          min_items=3) == [(0, 3), (3, 7)]
    assert split_list_for_multiprocessing([1, 2, 3, 4, 5, 6],
                                          max_folds=3,
                                          min_items=3) == [(0, 3), (3, 6)]
    assert split_list_for_multiprocessing([1, 2, 3, 4, 5, 6, 7],
                                          max_folds=3) == [(0, 2), (2, 4),
                                                           (4, 7)]


if __name__ == '__main__':
    start_processes(myinstance, ("abc", 123, [1, 5, 7]), num_processes=4)
    start_processes(myinstance1, (), num_processes=2)
    start_processes(myinstance2, ("test", ), num_processes=1)
    logger.info("End.")
    print(split_list_for_multiprocessing([1, 2, 3, 4, 5, 6, 7], max_folds=3))
    print(split_list_for_multiprocessing([1, 2, 3, 4, 5, 6], max_folds=3))
    print(split_list_for_multiprocessing([1, 2, 3, 4, 5, 6, 7], min_items=3))
    print(split_list_for_multiprocessing([1, 2, 3, 4, 5, 6], min_items=3))
    print(split_list_for_multiprocessing([1, 2, 3, 4, 5, 6, 7]))
    print(
        split_list_for_multiprocessing([1, 2, 3, 4, 5, 6, 7],
                                       max_folds=3,
                                       min_items=3))
    print(
        split_list_for_multiprocessing([1, 2, 3, 4, 5, 6],
                                       max_folds=3,
                                       min_items=3))
    print(split_list_for_multiprocessing([1, 2, 3, 4, 5, 6, 7], max_folds=3))

    test_split_list_for_multiprocessing_for_multiprocessing()



# -*- coding: utf-8 -*-
import os, re
import pandas as pd
import numpy as np
from multiprocessing import Process
from loguru import logger
from tqdm import tqdm

print(f"pwd: {os.path.abspath('.')}")
import sys
sys.path.append("seqlabel")
#  from tf_utils import enable_gpu_growth
from predict_config import Config
from data import load_train_data
from data import load_test_data, get_test_data_list
from seqlabel_model import BertModel, find_entities_from_indices
from seqlabel_utils import remove_head, remove_tail

sys.path.append(".")
from article import Section

debug_extract_labels = Config.debug_extract_labels


def post_process_base(s):
    s = s.replace("简历来自：BOSS直聘", '').replace("简历来自：BOSS直", '')
    return s


def post_process_PROJRESP(s):
    if len(s) <= 2:
        return ""
    if s[:1] in ['：', ':', '?']:
        s = s[1:]
    if s[:2] in ['至今']:
        s = s[2:]

    #  logger.debug(f"1 - post_process_PROJRESP: {s}")

    #  logger.debug(f"3 - post_process_PROJRESP: {s}")
    s = remove_head(s, "责任描述")

    #  logger.debug(f"6 - post_process_PROJRESP: {s}")
    re_string = "(\d{4}[\.\/\-\~]\d\d?\s*([—-][—-]?|至)\s*(\d{4}[\.\/\-\~]\d\d?|[至]?今)|\d{4}\s*年\s*\d\d?\s*月\s*[—-]\s*(\d{4}\s*年\s*\d\d?月))"

    m = re.search(re_string, s)
    if m:
        s0 = m.start()
        s = s[:s0]

    #  logger.debug(f"9 - post_process_PROJRESP: {s}")
    s = remove_tail(s, "项目描述")

    #  ignore_headers = [
    #      '该项目', '本项目', '该系统', '本系统', '此系统', '熟练使用', '公司刚开始', '虽然', '为三亚政府',
    #      '公司', '在校', '熟练', '火火免'
    #  ]
    #  for x in ignore_headers:
    #      if s[:len(x)] == x:
    #          return ""

    if s[:1] in ['：', ':', '?']:
        s = s[1:]
    #  logger.debug(f"post_process_PROJRESP: {s}")
    end_of_section = Section.end_of_section.replace(' ', '')
    if s[-len(end_of_section):] == end_of_section:
        s = s[:-len(end_of_section)]
    return s


def post_process_PNAME(s):
    if s[:4] in [
            '技术架构', '山地联盟', '信息化条', '京东金融', '银联前置', '软硬法规', '专题博物', 'Assi',
            '海南一禾', '基于国网', '华为PT', '利用二维', '个人电脑', '北京运通', '时代电气', '社会实践',
            '个人参与', '专业技能', '博通科技', '牛办是专'
    ]:
        return ""

    if s[:2] in ['至今']:
        s = s[2:]

    re_string = "(\d{4}[\.\/\-\~]\d\d?\s*([—-][—-]?|至)\s*(\d{4}[\.\/\-\~]\d\d?|[至]?今)|\d{4}\s*年\s*\d\d?\s*月\s*[—-]\s*(\d{4}\s*年\s*\d\d?月|至今))"

    m = re.search(re_string, s)
    if m:
        s0 = m.start()
        s = s[:s0]

    s = remove_tail(s, "，")
    s = remove_tail(s, "：")
    s = remove_tail(s, "、")
    s = remove_tail(s, ":")
    s = remove_tail(s, "|")
    s = remove_tail(s, "（")
    s = remove_tail(s, "()")
    s = remove_tail(s, "1")
    s = remove_tail(s, "一")
    s = remove_tail(s, "所属公司")
    s = remove_tail(s, "所属部门")
    s = remove_tail(s, "项目描述")
    s = remove_tail(s, "责任描述")
    s = remove_tail(s, "项目周期")
    s = remove_tail(s, "项目内容")
    s = remove_tail(s, "项目介绍")
    s = remove_tail(s, "学历本科")
    s = remove_tail(s, "职位")
    s = remove_tail(s, "UI设计师")
    s = remove_tail(s, "项目职责")
    s = remove_tail(s, "开发环境")
    s = remove_tail(s, "开发框架")
    s = remove_tail(s, "开发项目")
    s = remove_tail(s, "研发工程师")
    s = remove_tail(s, "使用软件")
    s = remove_tail(s, "主要技术")
    s = remove_tail(s, "软件开发平台")
    s = remove_tail(s, "责任心")
    s = remove_tail(s, "SSM框架")
    s = remove_tail(s, "独立完成")
    s = remove_tail(s, "运用python语言")
    s = remove_tail(s, "负责")

    return s


def post_process_WORKDUTY(s):
    i0 = s.find("工作内容")
    if i0 >= 0:
        s = s[:i0]
    return s


def post_process_WORKCONTENT(s):
    ignore_lines = ['到岗时间', '精通', 'C语言']
    for w in ignore_lines:
        if s.find(w) >= 0:
            return ""

    ignore_headers = [
        '本人', '为人', '熟练', '性格', '乐观', '开朗', '熟悉', '“负责任', '联系方式', '全、', '态度积极',
        '这段时间', '因为我从小'
    ]
    for x in ignore_headers:
        if s[:len(x)] == x:
            return ""
    s = remove_tail(s, "项目描述")

    end_of_section = Section.end_of_section.replace(' ', '')
    if s[-len(end_of_section):] == end_of_section:
        s = s[:-len(end_of_section)]

    return s


def post_process(s, category):
    s = post_process_base(s)
    if category == 'PROJRESP':
        s = post_process_PROJRESP(s)
    elif category == "WORKCONTENT":
        s = post_process_WORKCONTENT(s)
    elif category == "PNAME":
        s = post_process_PNAME(s)
    elif category == "WORKDUTY":
        s = post_process_WORKDUTY(s)
    #  logger.debug(f"post_process: {s}")
    return s


def do_extract_indices(test_data,
                       top_k=20,
                       s_threshold=0.01,
                       e_threshold=0.01,
                       min_label_len=8,
                       seg_len=510,
                       seg_backoff=400,
                       bert_model=None):
    if bert_model is None or type(bert_model) is str:
        if bert_model is None:
            from seqlabel_config import get_best_model_weights_file_name
            model_file = get_best_model_weights_file_name(selected_fold)
        elif type(bert_model) is str:
            #  model_file = f"./outputs/resumes_output/seqlabel/best_model_fold0.weights"
            model_file = bert_model

        bert_model = BertModel(Config.config_path,
                               Config.checkpoint_path,
                               Config.dict_path,
                               Config.learning_rate,
                               max_seq_length=seg_len)

        bert_model.build_model()

        bert_model.predict_model.load_weights(model_file)

    def _extract_indices(text_in, c_in, text_offset, top_k, s_threshold,
                         e_threshold):
        return bert_model.determine_entity_positions(text_in,
                                                     c_in,
                                                     text_offset,
                                                     top_k=top_k,
                                                     s_threshold=s_threshold,
                                                     e_threshold=e_threshold,
                                                     seg_len=seg_len,
                                                     seg_backoff=seg_backoff)

    def extract_indices_func():
        from functools import partial
        return partial(_extract_indices,
                       top_k=top_k,
                       s_threshold=s_threshold,
                       e_threshold=e_threshold)

    extract_indices = extract_indices_func()

    def output_results(test_data, extract_indices):
        indices = {}

        def add_to_indices(ID, text_in, c_in, data_indices):
            if data_indices:
                if ID not in indices:
                    indices[ID] = []
                indices[ID].append([text_in, c_in, data_indices])

        # 解决最后一项抽取不出来的问题
        #  end_of_section = "--- End of section ---"
        n = 0
        for ID, text_in, c_in, text_offset in tqdm(
                test_data, desc="SeqLabels predicting"):

            #  need_end_of_section = False
            #  if len(test_data) == 1:
            #      need_end_of_section = True
            #  elif n == len(test_data) - 1:
            #      need_end_of_section = True
            #  elif n < len(test_data) - 1:
            #      next_ID = test_data[n + 1][0]
            #      if next_ID != ID:
            #          need_end_of_section = True
            #  if need_end_of_section:
            #      text_in += end_of_section
            #      if len(text_in) > seg_len:
            #          #  logger.info(f"{ID}: {text_in}")
            #          data_indices = extract_indices(text_in[:seg_len], c_in,
            #                                         text_offset)
            #          add_to_indices(ID, text_in[:seg_len], c_in, data_indices)
            #
            #          text_in = text_in[seg_len - seg_backoff:]
            #          text_offset += seg_len - seg_backoff
            #
            #          #  logger.warning(f"{ID}: {text_in}")

            data_indices = extract_indices(text_in, c_in, text_offset)
            #  logger.debug(f"{data_indices}")
            #  logger.debug(f"{text_in}")
            add_to_indices(ID, text_in, c_in, data_indices)
            n += 1

        return indices

    def resort_indices(indices):
        new_indices = {}
        for ID, X in indices.items():
            Y = [(text_in, c_in, (t, idx, prob, text_offset))
                 for text_in, c_in, data_indices in X
                 for t, idx, prob, text_offset in data_indices]
            Y = sorted(Y, key=lambda x: x[2][1] + x[2][3])
            new_indices[ID] = Y
        return new_indices

    indices = output_results(test_data, extract_indices)
    indices = resort_indices(indices)

    return indices


#  def do_extract_labels(test_data,
#                        top_k=20,
#                        s_threshold=0.01,
#                        e_threshold=0.01,
#                        min_label_len=8):
#      bert_model = BertModel(Config.config_path, Config.checkpoint_path,
#                             Config.dict_path, Config.learning_rate)
#
#      bert_model.build_model()
#
#      def _extract_entity(text_in, c_in, top_k, s_threshold, e_threshold,
#                          min_label_len):
#          return bert_model.extract_entities(text_in,
#                                             c_in,
#                                             top_k=top_k,
#                                             s_threshold=s_threshold,
#                                             e_threshold=e_threshold,
#                                             min_label_len=min_label_len)
#
#      def extract_entity_func():
#          from functools import partial
#          return partial(_extract_entity,
#                         top_k=top_k,
#                         s_threshold=s_threshold,
#                         e_threshold=e_threshold,
#                         min_label_len=min_label_len)
#
#      extract_entity = extract_entity_func()
#
#      model_file = get_best_model_weights_file_name(selected_fold)
#      #  model_file = f"./outputs/resumes_output/seqlabel/best_model_fold0.weights"
#      bert_model.predict_model.load_weights(model_file)
#
#      def output_results(test_data, extract_entity):
#          results = {}
#
#          def add_to_results(ID, label_text):
#              if label_text:
#                  if ID not in results:
#                      results[ID] = []
#                  results[ID].append(label_text)
#
#          last_entity_texts = []
#          last_entities = []
#          last_ID = None
#          last_dangling_entity = None
#          #  for d in tqdm(iter(test_data)):
#          for d in tqdm(test_data, desc="SeqLabels predicting"):
#              ID = d[0]
#
#              # 处理悬垂实体
#              if ID != last_ID:
#                  if last_ID is not None:
#                      if last_dangling_entity:
#                          idx0, prob0, entity_text = last_dangling_entity
#                          add_to_results(last_ID, entity_text)
#                          #  s = f"{last_ID}【Diangling({idx0}:{idx0+len(entity_text)}), prob: {prob0:.6f})】,{entity_text}\n"
#                          #  F.write(s)
#
#                  #  F.write(
#                  #      "\n============================================================\n"
#                  #  )
#              entities, dangling_entity = extract_entity(d[1], d[2])
#
#              selected_entities = 0
#              for entity_text, idx0, idx1, prob0, prob1 in entities:
#                  # FIX, 重叠窗口重复抽取的问题
#                  if ID == last_ID and entity_text in last_entity_texts:
#                      #  if ID == last_ID and idx0 < seg_backoff and idx1 < seg_backoff:
#                      #  F.write(
#                      #      f"!!!!!! ({idx0}:{idx1})({prob0:.6f}:{prob1:.6f} --------> skip {entity_text}"
#                      #  )
#                      continue
#
#                  add_to_results(ID, entity_text)
#                  #  s = f"{ID}【({idx0}:{idx1})({prob0:.6f}:{prob1:.6f})】,{entity_text}\n"
#                  #  #  s = f"{ID},{entity_text}\n"
#                  #  F.write(s)
#                  selected_entities += 1
#              #  if selected_entities == 0:
#              #      F.write("........................................\n")
#              #  else:
#              #      F.write("----------------------------------------\n")
#              last_entity_texts = [
#                  entity_text for entity_text, _, _, _, _ in entities
#              ]
#
#              last_ID = ID
#              last_dangling_entity = dangling_entity
#          return results
#
#      results = output_results(test_data, extract_entity)
#      return results


def seg_long_text(text, seg_len, seg_backoff):
    texts = []
    n_chars = len(text)
    if n_chars > 0:
        s = 0
        while s < n_chars:
            seg_text = text[s:s + seg_len]
            texts.append(seg_text)
            if s + seg_len >= n_chars:
                break
            s += seg_len - seg_backoff
    return texts


def extract_labels(resume_texts,
                   category,
                   top_k=20,
                   s_threshold=0.01,
                   e_threshold=0.01,
                   min_label_len=8,
                   seg_len=510,
                   seg_backoff=400,
                   bert_model=None):
    test_data = []
    for ID, text in resume_texts.items():

        content_text = text
        content_text = content_text.replace(' ', '')
        if content_text:
            seg_texts = seg_long_text(content_text, seg_len, seg_backoff)
            for seg_text in seg_texts:
                test_data.append([ID, seg_text, category])
        #  test_data.append((ID, text, category))

    #  labels = do_extract_labels(test_data,
    #                             top_k=top_k,
    #                             s_threshold=s_threshold,
    #                             e_threshold=e_threshold,
    #                             min_label_len=min_label_len)
    results = extract_labels_from_test_data(test_data,
                                            category,
                                            top_k=top_k,
                                            s_threshold=s_threshold,
                                            e_threshold=e_threshold,
                                            min_label_len=min_label_len,
                                            seg_len=seg_len,
                                            seg_backoff=seg_backoff,
                                            bert_model=bert_model)

    #  print(results)
    labels = {ID: [x[0] for x in items] for ID, items in results.items()}
    #  print(labels)
    return labels


def adjust_test_data(test_data, seg_len, seg_backoff):
    tmp_data = {}
    for ID, text_in, c_in in test_data:
        if ID not in tmp_data:
            tmp_data[ID] = []
        tmp_data[ID].append((text_in, c_in))
    new_test_data = []
    for ID, X in tmp_data.items():
        X = [(ID, text_in, c_in, (seg_len - seg_backoff) * i)
             for i, (text_in, c_in) in enumerate(X)]
        new_test_data += X

    return new_test_data


def extract_labels_from_test_data(test_data,
                                  category,
                                  top_k=20,
                                  s_threshold=0.01,
                                  e_threshold=0.01,
                                  min_label_len=8,
                                  seg_len=510,
                                  seg_backoff=400,
                                  bert_model=None):

    test_data = adjust_test_data(test_data, seg_len, seg_backoff)
    indices = do_extract_indices(test_data,
                                 top_k=top_k,
                                 s_threshold=s_threshold,
                                 e_threshold=e_threshold,
                                 seg_len=seg_len,
                                 seg_backoff=seg_backoff,
                                 bert_model=bert_model)

    full_texts = {}
    for ID, text, category, text_offset in test_data:
        if ID not in full_texts:
            full_texts[ID] = []
        full_texts[ID].append(text)

    for ID, texts in full_texts.items():
        full_text = ""
        for text_in in texts:
            full_text += text_in[:seg_len - seg_backoff]
        if texts:
            full_text += texts[-1][seg_len - seg_backoff:]
        #  logger.debug(f"{full_text}")
        full_texts[ID] = full_text

    if debug_extract_labels:
        for ID, data_indices in indices.items():
            logger.info(f"-------- {ID} --------")
            for text_in, c_in, (tag, idx, prob, off) in data_indices:
                logger.debug(
                    f"{tag}, {idx}, {prob:.3f}, {off}, {idx+off} - {full_texts[ID][idx+off:idx+off+10]}"
                )

    # -----------------------------------------------------------------
    results = {}

    def add_to_results(ID, label_text, idx0, idx1, prob0, prob1):
        if label_text:
            if ID not in results:
                results[ID] = []
            results[ID].append((label_text, idx0, idx1, prob0, prob1))

    last_entity_texts = []
    last_entities = []
    last_ID = None
    last_dangling_entity = None
    last_full_text = ""
    #  for ID, [text_in, c_in, data_indices] in indices.items():
    for ID, data_indices in tqdm(indices.items(), desc="Extract entities:"):
        full_text = full_texts.get(ID, "")
        if debug_extract_labels:
            logger.info(f"full_text: {full_text}")
        entities, dangling_entity = find_entities_from_indices(
            data_indices, min_label_len)

        if debug_extract_labels:
            logger.info(
                f"entities: {entities}, dangling_entity: {dangling_entity}")

        # 处理悬垂实体
        if ID != last_ID:
            if last_ID is not None:
                if last_dangling_entity:
                    idx0, prob0, entity_text, c_in = last_dangling_entity

                    idx1 = len(last_full_text)
                    entity_text = last_full_text[idx0:idx1]
                    #  logger.debug(
                    #      f"diangling entity({idx0}:{idx1}): {entity_text}")
                    entity_text = post_process(entity_text, category)
                    if debug_extract_labels:
                        logger.warning(
                            f"add diangling entity({idx0}:{idx1}): {entity_text}"
                        )
                    if len(entity_text) > 2:
                        add_to_results(last_ID, entity_text, idx0, idx1, prob0,
                                       0.0)
                    #  s = f"{last_ID}【Diangling({idx0}:{idx0+len(entity_text)}), prob: {prob0:.6f})】,{entity_text}\n"
                    #  logger.debug(s)
                    #  F.write(s)

            #  logger.debug(f"\n============================================================\n")

        selected_entities = 0
        for idx0, idx1, prob0, prob1, entity_text, c_in in entities:
            # FIX, 重叠窗口重复抽取的问题
            if ID == last_ID and entity_text in last_entity_texts:
                #  s = f"  !!!!!! ({idx0}:{idx1})({prob0:.6f}:{prob1:.6f} --------> skip {entity_text}"
                #  logger.debug(s)
                continue

            entity_text = full_text[idx0:idx1]
            #  logger.debug(f"entity({idx0}:{idx1}): {entity_text}")
            entity_text = post_process(entity_text, category)
            if debug_extract_labels:
                logger.warning(f"add entity({idx0}:{idx1}): {entity_text}")
            if len(entity_text) > 2:
                add_to_results(ID, entity_text, idx0, idx1, prob0, prob1)
            #  s = f"{ID}【({idx0}:{idx1})({prob0:.6f}:{prob1:.6f})】,{entity_text}\n"
            #  logger.debug(s)
            selected_entities += 1
        #  if selected_entities == 0:
        #      logger.debug("........................................\n")
        #      #  F.write("........................................\n")
        #  else:
        #      logger.debug("----------------------------------------\n")
        last_entity_texts = [
            entity_text for _, _, _, _, entity_text, c_in in entities
        ]
        last_full_text = full_text

        last_ID = ID
        last_dangling_entity = dangling_entity

    if last_ID and last_dangling_entity:
        idx0, prob0, entity_text, c_in = last_dangling_entity

        idx1 = len(last_full_text)
        entity_text = last_full_text[idx0:idx1]

        entity_text = post_process(entity_text, category)
        if debug_extract_labels:
            logger.warning(
                f"add diangling entity({idx0}:{idx1}): {entity_text}")
        if len(entity_text) > 2:
            add_to_results(last_ID, entity_text, idx0, idx1, prob0, 0.0)
        #  s = f"{last_ID}【Diangling({idx0}:{idx0+len(entity_text)}), prob: {prob0:.6f})】,{entity_text}\n"
        #  logger.debug(s)
        #  F.write(s)

    return results


#  def predict_one_fold(test_data, selected_fold, overwrite_result_file=False):
#      from seqlabel_config import get_result_file_name
#      output_file = get_result_file_name(selected_fold)
#      if not overwrite_result_file and os.path.isfile(output_file):
#          print(f"{output_file} exist.")
#          return
#
#      labels = extract_labels_from_test_data(test_data,
#                                             category=Config.category,
#                                             top_k=Config.top_k,
#                                             s_threshold=Config.s_threshold,
#                                             e_threshold=Config.e_threshold,
#                                             min_label_len=Config.min_label_len,
#                                             seg_len=Config.seg_len,
#                                             seg_backoff=Config.seg_backoff)
#
#      with open(output_file, 'w') as F:
#          for ID, entities in labels.items():
#              for entity_text, idx0, idx1, prob0, prob1 in entities:
#                  s = f"{ID}【({idx0:04d}:{idx1:04d})({prob0:.6f}:{prob1:.6f})】| {entity_text}\n"
#                  F.write(s)
#              F.write("\n")
#


def get_args():
    import argparse
    parser = argparse.ArgumentParser()

    parser.add_argument("--fold",
                        type=int,
                        default=-0,
                        help="Special fold model used for predicting")
    parser.add_argument("--overwrite_result_file",
                        action="store_true",
                        help="Overwrite the content of the result file")

    args = parser.parse_args()

    return args


if __name__ == '__main__':

    args = get_args()

    selected_fold = args.fold

    np.random.seed(Config.SEED)

    logger.add(f"log/test_seqlabel_{Config.experiment}_{Config.category}.log")

    if not os.path.exists(Config.results_dir):
        os.makedirs(Config.results_dir)

    #  df_test = load_test_data(Config.raw_test_data_file, Config.reverse_text)
    #  test_data = get_test_data_list(df_test)
    #  logger.info(f"selected fold: {selected_fold}")

    #  predict_one_fold(test_data, selected_fold, args.overwrite_result_file)

    #  batch_folds = 1
    #  total_folds = Config.total_folds
    #  folds = [total_folds - x - 1 for x in range(total_folds)]
    #  while len(folds) > 0:
    #      selected_folds = []
    #      for i in range(min(batch_folds, len(folds))):
    #          selected_folds.append(folds.pop())
    #      for selected_fold in selected_folds:
    #
    #          test_data = get_test_data_list(df_test)
    #
    #          logger.info(f"selected fold: {selected_fold}")
    #
    #          #  process = Process(target=predict_one_fold_old,
    #          #                    args=(test_data, selected_fold, classes, True))
    #          process = Process(target=predict_one_fold,
    #                            args=(test_data, selected_fold, classes, True))
    #          process.start()
    #      process.join()
    #      break



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from loguru import logger


def get_ner_results(metric):
    eval_info, entity_info = metric.result()
    results = {f'{key}': value for key, value in eval_info.items()}

    title = f"{' '*32}   acc    recall f1    "
    title_len = len(title.encode('gbk'))
    logger.info(f"{'=' * title_len}")
    logger.info(title)
    logger.info(f"{'-' * title_len}")

    sorted_entity_info = sorted(entity_info.items(),
                                key=lambda x: x[1]['f1'],
                                reverse=True)
    for key, metrics in sorted_entity_info:
        #  for key in entity_info.keys():
        #      metrics = entity_info[key]
        disp_key = key[:16]
        disp_key += ' ' * (32 - len(disp_key.encode('gbk')))
        info = f"{disp_key} | {metrics['acc']:.4f} {metrics['recall']:.4f} {metrics['f1']:.4f}"
        logger.info(info)
    logger.info(f"{'-' * title_len}")
    info = f"{' '*32} | {results['acc']:.4f} {results['recall']:.4f} {results['f1']:.4f}"  #" - loss: {results['loss']:.4f}"
    logger.info(info)
    logger.info(f"{'-' * title_len}")

    return results




#!/usr/bin/env python
# -*- coding: utf-8 -*-

import numpy as np
from loguru import logger

import torch
import torch.nn as nn
from torch.nn import CrossEntropyLoss
import torch.nn.functional as F
#  from ..models.crf_0 import CRF as CRF0
#  from ..models.crf import CRF

#  from ..models.ncrfpp_crf import CRF as CRFPP

#  from ..models.sltk_crf import CRF as CRFSLTK

#  from TorchCRF import CRF

#  from torchcrf import CRF
# https://github.com/kmkurn/pytorch-crf/blob/master/torchcrf/__init__.py

#  from ..models.lstm_crf import CRF

#  from .models.bert_for_ner import BertCrfForNer
from .utils import CNerTokenizer, SeqEntityScore, get_entities
from ..trainer import Trainer, get_default_optimizer_parameters
from ...utils.multiprocesses import barrier_leader_process, barrier_member_processes, is_multi_processes

from torch.utils.data import DataLoader, RandomSampler, SequentialSampler
from transformers import AutoConfig, BertTokenizerFast, AutoModelForTokenClassification
from transformers.modeling_bert import BertPreTrainedModel, BertModel

CRF_TYPE = "old_crf"
CRF_TYPE = "new_crf"
CRF_TYPE = "lstm_crf"
CRF_TYPE = "TorchCRF"
CRF_TYPE = "pytorch-crf"
CRF_TYPE = "ncrfpp"
CRF_TYPE = "crfsltk"


class CRFLayer:
    def __init__(self, tagset_size, start_tag_idx, stop_tag_idx):
        self.tagset_size = tagset_size

    def get_loss(self, logits, labels, mask=None):
        raise NotImplementedError

    def decode(self, logits, mask=None):
        raise NotImplementedError


class OldCRF(CRFLayer):
    def __init__(self, tagset_size, start_tag_idx, stop_tag_idx):
        from ..models.crf_0 import CRF as CRF0
        self.crf = CRF0(tagset_size, start_tag_idx, stop_tag_idx)

    def get_loss(self, logits, labels, input_lens=None, mask=None):
        loss = self.crf.calculate_loss(logits, labels, lengths=input_lens)
        return loss

    def decode(self, logits, input_lens=None, mask=None):
        seq_path, _ = self.crf.obtain_labels(logits, input_lens)
        return seq_path


class NewCRF(CRFLayer):
    def __init__(self, tagset_size, start_tag_idx, stop_tag_idx):
        from ..models.crf import CRF
        self.crf = CRF(tagset_size, start_tag_idx, stop_tag_idx)

    def get_loss(self, logits, labels, mask=None):
        loss = self.crf(logits, labels)
        # batch_second
        #  loss = self.crf(logits.transpose(0, 1), labels.transpose(0, 1))
        return loss

    def decode(self, logits, mask=None):
        seq_path = self.crf.decode(logits, mask)
        seq_path = seq_path.cpu().numpy().tolist()[0]
        return seq_path


class NCRFPP(CRFLayer):
    def __init__(self, tagset_size, start_tag_idx, stop_tag_idx):
        from ..models.ncrfpp_crf import CRF as CRFPP
        self.crf = CRFPP(tagset_size, start_tag_idx, stop_tag_idx)

    def get_loss(self, logits, labels, mask=None):
        loss = self.crf.neg_log_likelihood_loss(logits, mask, labels)
        return loss

    def decode(self, logits, mask=None):
        scores, seq_path = self.crf._viterbi_decode(logits, mask)
        seq_path = seq_path.cpu().numpy().tolist()
        return seq_path


class CRFSLTK(CRFLayer):
    def __init__(self, tagset_size, start_tag_idx, stop_tag_idx):
        from ..models.sltk_crf import CRF as CRFSLTK
        self.crf = CRFSLTK(target_size, start_tag_idx, stop_tag_idx)

    def get_loss(self, logits, labels, mask=None):
        loss = self.crf.neg_log_likelihood_loss(logits, mask.bool(), labels)
        return loss

    def decode(self, logits, mask=None):
        _, seq_path = model.crf(logits, mask.bool())
        return seq_path


class TorchCRF(CRFLayer):
    def __init__(self, tagset_size, start_tag_idx, stop_tag_idx):
        from TorchCRF import CRF
        self.crf = CRF(tagset_size)

    def get_loss(self, logits, labels, mask=None):
        loss = self.crf(logits, labels, mask.bool())
        return loss

    def decode(self, logits, mask=None):
        seq_path = model.crf.viterbi_decode(logits, mask.bool())
        return seq_path


class PyTorchCRF(CRFLayer):
    def __init__(self, tagset_size, start_tag_idx, stop_tag_idx):
        from torchcrf import CRF
        # https://github.com/kmkurn/pytorch-crf/blob/master/torchcrf/__init__.py
        self.crf = CRF(tagset_size, batch_first=torch.BoolTensor([True]))

    def get_loss(self, logits, labels, mask=None):
        #  # preds = np.argmax(logits.cpu().detach().numpy(), axis=2)
        #  # preds = torch.argmax(logits, axis=2)
        loss = -self.crf(logits, labels)  #, mask)
        return loss

    def decode(self, logits, mask=None):
        seq_path = self.crf.decode(logits)  #, mask.bool())
        return seq_path


class LSTMCRF(CRFLayer):
    def __init__(self, tagset_size, start_tag_idx, stop_tag_idx):
        from ..models.lstm_crf import CRF
        self.crf = CRF(tagset_size, start_tag_idx, stop_tag_idx)

    def get_loss(self, logits, labels, mask=None):
        loss = self.crf.neg_log_likelihood(logits, labels)
        return loss

    def decode(self, logits, mask=None):
        seq_path = model.crf(logits)
        return seq_path


class BertCrfForNer(BertPreTrainedModel):
    def __init__(self, config):
        super(BertCrfForNer, self).__init__(config)
        #  self.num_labels = len(label2id)

        self.num_labels = config.num_labels
        self.crf_type = config.crf_type
        self.no_crf_loss = config.no_crf_loss

        self.bert = BertModel(config)
        self.dropout = nn.Dropout(config.hidden_dropout_prob)
        self.classifier = nn.Linear(config.hidden_size, self.num_labels)

        label2id = config.label2id
        start_tag_idx = label2id["[CLS]"]
        stop_tag_idx = label2id["[SEP]"]

        tagset_size = len(label2id)

        # CRFPP
        if self.crf_type == 'ncrfpp':
            from ..models.ncrfpp_crf import CRF as CRFPP
            self.crf = CRFPP(tagset_size, start_tag_idx, stop_tag_idx)
        elif self.crf_type == 'pytorch-crf':
            from torchcrf import CRF
            # https://github.com/kmkurn/pytorch-crf/blob/master/torchcrf/__init__.py
            self.crf = CRF(len(label2id), batch_first=torch.BoolTensor([True]))
            #  self.crf = PyTorchCRF(tagset_size, start_tag_idx, stop_tag_idx)
        elif self.crf_type == 'old_crf':
            from ..models.crf_0 import CRF
            #  self.crf_0 = CRF0(tagset_size=len(label2id),
            #                    tag_dictionary=label2id,
            #                    device=device)
            self.crf_0 = CRF(tagset_size, start_tag_idx, stop_tag_idx)
            #  self.crf = OldCRF(tagset_size, start_tag_idx, stop_tag_idx)
        elif self.crf_type == 'new_crf':
            from ..models.crf import CRF
            self.crf = CRF(len(label2id), batch_first=True)
            # self.crf = NewCRF(tagset_size, start_tag_idx, stop_tag_idx)
        elif self.crf_type == 'lstm_crf':
            from ..models.lstm_crf import CRF
            self.crf = CRF(tagset_size, start_tag_idx, stop_tag_idx)
            #  self.crf = LSTMCRF(tagset_size, start_tag_idx, stop_tag_idx)

        # lstm_crf
        #  self.crf = CRF(tagset_size, start_tag_idx, stop_tag_idx)
        #  self.crf = LSTMCRF(tagset_size, start_tag_idx, stop_tag_idx)

        # TorchCRF
        #  self.crf = CRF(tagset_size, start_tag_idx, stop_tag_idx)
        #  self.crf = TorchCRF(tagset_size, start_tag_idx, stop_tag_idx)

        # torchcrf
        #  self.crf = CRF(len(label2id), batch_first=torch.BoolTensor([True]))
        #  self.crf = PyTorchCRF(tagset_size, start_tag_idx, stop_tag_idx)

        # NewCRF
        #  self.crf = CRF(len(label2id), batch_first=True)
        # self.crf = NewCRF(tagset_size, start_tag_idx, stop_tag_idx)

        #  self.crf_0 = CRF0(tagset_size=len(label2id),
        #                    tag_dictionary=label2id,
        #                    device=device)
        #  self.crf_0 = CRF0(tagset_size, start_tag_idx, stop_tag_idx)
        #  self.crf = OldCRF(tagset_size, start_tag_idx, stop_tag_idx)

        # CRFPP
        #  self.crf = CRFPP(tagset_size, start_tag_idx, stop_tag_idx)
        #  self.crf = NCRFPP(tagset_size, start_tag_idx, stop_tag_idx)

        # CRFSLTK
        #  self.crf = CRFSLTK(target_size, start_tag_idx, stop_tag_idx)
        #  self.crf = CRFSLTK(tagset_size, start_tag_idx, stop_tag_idx)

        self.init_weights()

    def forward(self,
                input_ids,
                attention_mask=None,
                token_type_ids=None,
                labels=None,
                input_lens=None):
        outputs = self.bert(input_ids, token_type_ids, attention_mask)
        sequence_output = outputs[0]
        sequence_output = self.dropout(sequence_output)
        logits = self.classifier(sequence_output)
        outputs = (logits, )
        if labels is not None:
            loss_fct = CrossEntropyLoss()
            # Only keep active parts of the loss
            if attention_mask is not None:
                active_loss = attention_mask.view(-1) == 1
                active_logits = logits.view(-1, self.num_labels)
                active_labels = torch.where(
                    active_loss, labels.view(-1),
                    torch.tensor(loss_fct.ignore_index).type_as(labels))
                clf_loss = loss_fct(active_logits, active_labels)
            else:
                clf_loss = loss_fct(logits.view(-1, self.num_labels),
                                    labels.view(-1))

            crf_loss = clf_loss

            if self.no_crf_loss:
                loss = clf_loss
            else:
                if self.crf_type == 'ncrfpp':
                    crf_loss = self.crf.neg_log_likelihood_loss(
                        logits, attention_mask, labels)
                elif self.crf_type == 'pytorch-crf':
                    # pytorch-crf
                    crf_loss = -self.crf(logits, labels)  #, attention_mask)
                elif self.crf_type == 'old_crf':
                    crf_loss = self.crf.calculate_loss(logits,
                                                       tag_list=labels,
                                                       lengths=input_lens)
                elif self.crf_type == 'new_crf':
                    crf_loss = self.crf(logits, labels)
                    #  crf_loss = self.crf(logits.transpose(0, 1), labels.transpose(0, 1))
                elif self.crf_type == 'lstm_crf':
                    crf_loss = self.crf.neg_log_likelihood(logits, labels)

                loss = crf_loss

            #  crf_loss = self.crf.get_loss(logits, labels, attention_mask)

            # lstm_crf
            #  crf_loss = self.crf.neg_log_likelihood(logits, labels)

            # TorchCRF
            #  crf_loss = self.crf(logits, labels, attention_mask.bool())

            # torchcrf
            #  # preds = np.argmax(logits.cpu().detach().numpy(), axis=2)
            #  # preds = torch.argmax(logits, axis=2)
            #  crf_loss = -self.crf(logits, labels)  #, attention_mask)

            #  crf_loss = self.crf_0.calculate_loss(logits,
            #                                       tag_list=labels,
            #                                       lengths=input_lens)

            # CRFPP
            #  crf_loss = self.crf.neg_log_likelihood_loss(
            #      logits, attention_mask, labels)

            # new local crf
            #  crf_loss = self.crf(logits, labels)
            #  crf_loss = self.crf(logits.transpose(0, 1), labels.transpose(0, 1))

            # CRFSLTK
            #  crf_loss = self.crf.neg_log_likelihood_loss(
            #      logits, attention_mask.bool(), labels)

            #  crf_loss = 0.0

            #  loss = clf_loss + crf_loss
            #  loss = crf_loss
            outputs = (loss, ) + outputs
            return outputs
        else:
            #  return (0.0, ) + outputs  # (loss), scores
            #  return (torch.tensor(0.0).cuda(), ) + outputs
            return outputs


#  from transformers import AlbertConfig
#  from transformers.modeling_albert import AlbertPreTrainedModel, AlbertModel
#
#
#  class AlbertCrfForNer(AlbertPreTrainedModel):
#      def __init__(self, config, label2id, device):
#          super(AlbertCrfForNer, self).__init__(config)
#          self.bert = AlbertModel(config)
#          self.dropout = nn.Dropout(config.hidden_dropout_prob)
#          self.classifier = nn.Linear(config.hidden_size, len(label2id))
#          self.crf = CRF(tagset_size=len(label2id),
#                         tag_dictionary=label2id,
#                         device=device,
#                         is_bert=True)
#          self.init_weights()
#
#      def forward(self,
#                  input_ids,
#                  token_type_ids=None,
#                  attention_mask=None,
#                  labels=None,
#                  input_lens=None):
#          outputs = self.bert(input_ids, token_type_ids, attention_mask)
#          sequence_output = outputs[0]
#          sequence_output = self.dropout(sequence_output)
#          logits = self.classifier(sequence_output)
#          outputs = (logits, )
#          if labels is not None:
#              loss = self.crf.calculate_loss(logits,
#                                             tag_list=labels,
#                                             lengths=input_lens)
#              outputs = (loss, ) + outputs
#          return outputs  # (loss), scores
#
#
#  from transformers import XLNetConfig
#  from transformers.modeling_xlnet import XLNetPreTrainedModel, XLNetModel
#
#
#  class XLNetCrfForNer(XLNetPreTrainedModel):
#      def __init__(self, config, label2id, device):
#          super(XLNetCrfForNer, self).__init__(config)
#          self.bert = XLNetModel(config)
#          #  self.dropout = nn.Dropout(config.hidden_dropout_prob)
#          self.dropout = nn.Dropout(0.5)
#          self.classifier = nn.Linear(config.hidden_size, len(label2id))
#          self.crf = CRF(tagset_size=len(label2id),
#                         tag_dictionary=label2id,
#                         device=device,
#                         is_bert=True)
#          self.init_weights()
#
#      def forward(self,
#                  input_ids,
#                  token_type_ids=None,
#                  attention_mask=None,
#                  labels=None,
#                  input_lens=None):
#          outputs = self.bert(input_ids, token_type_ids, attention_mask)
#          sequence_output = outputs[0]
#          sequence_output = self.dropout(sequence_output)
#          logits = self.classifier(sequence_output)
#          outputs = (logits, )
#          if labels is not None:
#              loss = self.crf.calculate_loss(logits,
#                                             tag_list=labels,
#                                             lengths=input_lens)
#              outputs = (loss, ) + outputs
#          return outputs  # (loss), scores

MODEL_CLASSES = {
    'bert': (AutoConfig, BertCrfForNer, CNerTokenizer),
    #  'bert': (AutoConfig, AutoModelForTokenClassification, CNerTokenizer),
    #  'albert': (AlbertConfig, AlbertCrfForNer, CNerTokenizer),
    #  'xlnet': (XLNetConfig, XLNetCrfForNer, CNerTokenizer),
}


def load_pretrained_tokenizer(args):
    config_class, model_class, tokenizer_class = MODEL_CLASSES[args.model_type]
    tokenizer = tokenizer_class.from_pretrained(
        args.model_path,
        do_lower_case=args.do_lower_case,
        cache_dir=args.cache_dir if args.cache_dir else None,
    )

    return tokenizer


def load_pretrained_model(args):
    # make sure only the first process in distributed training
    # will download model & vocab
    barrier_member_processes(args)

    config_class, model_class, tokenizer_class = MODEL_CLASSES[args.model_type]
    config = config_class.from_pretrained(
        args.model_path,
        cache_dir=args.cache_dir if args.cache_dir else None,
        num_labels=args.num_labels,
        label2id=args.label2id,
    )
    setattr(config, 'crf_type', args.crf_type)
    setattr(config, 'no_crf_loss', args.no_crf_loss)

    logger.info(f"model_path: {args.model_path}")
    logger.info(f"config:{config}")
    model = model_class.from_pretrained(
        args.model_path,
        from_tf=bool(".ckpt" in args.model_path),
        config=config,
        cache_dir=args.cache_dir if args.cache_dir else None,
    )

    # make sure only the first process in distributed training
    # will download model & vocab
    barrier_leader_process(args)

    return model


#  def load_pretrained_tokenizer(args):
#      from .trainer import load_pretrained_tokenizer as load_pretrained_tokenizer_base
#      return load_pretrained_tokenizer_base(args, MODEL_CLASSES)
#
#
#  def load_pretrained_model(args):
#      from .trainer import load_pretrained_model as load_pretrained_model_base
#      return load_pretrained_model_base(args, MODEL_CLASSES)
#


def collate_fn(batch):
    """
    batch should be a list of (sequence, target, length) tuples...
    Returns a padded tensor of sequences sorted from longest to shortest,
    """
    all_input_ids, all_attention_mask, all_token_type_ids, all_labels, all_lens = map(
        torch.stack, zip(*batch))
    max_len = max(all_lens).item()
    all_input_ids = all_input_ids[:, :max_len]
    all_attention_mask = all_attention_mask[:, :max_len]
    all_token_type_ids = all_token_type_ids[:, :max_len]
    all_labels = all_labels[:, :max_len]
    return all_input_ids, all_attention_mask, all_token_type_ids, all_labels, all_lens


def load_model(args):
    model = load_pretrained_model(args)
    model.to(args.device)
    return model


def build_default_model(args):
    """
    自定义模型
    规格要求返回模型(model)、优化器(optimizer)、调度器(scheduler)三元组。
    """

    # -------- model --------
    model = load_pretrained_model(args)
    model.to(args.device)

    # -------- optimizer --------
    from transformers.optimization import AdamW
    optimizer_parameters = get_default_optimizer_parameters(
        model, args.weight_decay)
    optimizer = AdamW(optimizer_parameters,
                      lr=args.learning_rate,
                      correct_bias=False)

    # -------- scheduler --------
    from transformers.optimization import get_linear_schedule_with_warmup

    scheduler = get_linear_schedule_with_warmup(
        optimizer,
        num_warmup_steps=args.total_steps * args.warmup_rate,
        num_training_steps=args.total_steps)

    return model, optimizer, scheduler


def preds_to_entities(preds, id2label, markup='bios', autofix=False):
    #  logger.debug(f"{preds}")
    preds = [id2label[x] for x in preds]
    label_entities = get_entities(preds,
                                  id2label=id2label,
                                  markup=markup,
                                  autofix=autofix)
    return label_entities


def init_labels(args, ner_labels):
    label_set = ['X', 'O', '[CLS]', '[SEP]']
    for x in ner_labels:
        label_set += [f"B-{x}", f"I-{x}", f"S-{x}"]
    args.id2label = {i: label for i, label in enumerate(label_set)}
    args.label2id = {label: i for i, label in enumerate(label_set)}
    args.num_labels = len(label_set)

    logger.info(f"args.label2id: {args.label2id}")
    logger.info(f"args.id2label: {args.id2label}")
    logger.info(f"args.num_labels: {args.num_labels}")


class NerTrainer(Trainer):
    def __init__(self, args, ner_labels, build_model=None, tokenizer=None):
        super(NerTrainer, self).__init__(args)
        init_labels(args, ner_labels)
        if tokenizer:
            self.tokenizer = tokenizer
        else:
            self.tokenizer = load_pretrained_tokenizer(args)

        if build_model is None:
            self.build_model = build_default_model
        else:
            self.build_model = build_model

        self.label2id = args.label2id
        self.collate_fn = collate_fn

    def batch_to_inputs(self, args, batch, known_labels=True):
        inputs = {
            "input_ids": batch[0],
            "attention_mask": batch[1],
            "labels": batch[3] if known_labels else None,
            'input_lens': batch[4]
        }
        if args.model_type != "distilbert":
            # XLM and RoBERTa don"t use segment_ids
            inputs["token_type_ids"] = (
                batch[2] if args.model_type in ["bert", "xlnet"] else None)
        return inputs

    def examples_to_dataset(self, examples, max_seq_length):
        from .dataset import examples_to_dataset
        return examples_to_dataset(examples, self.label2id, self.tokenizer,
                                   max_seq_length)

    #  def generate_dataloader(self, args, dataset, batch_size, keep_order=True):
    #
    #      Sampler = SequentialSampler if keep_order else RandomSampler
    #      sampler = DistributedSampler(dataset) if is_multi_processes(
    #          args) else Sampler(dataset)
    #      dataloader = DataLoader(dataset,
    #                              sampler=sampler,
    #                              batch_size=batch_size,
    #                              collate_fn=collate_fn)
    #      return dataloader

    def on_eval_start(self, args, eval_dataset):
        self.metric = SeqEntityScore(args.id2label,
                                     markup=args.markup,
                                     autofix=args.autofix)
        self.error_entities = []

    #  def on_eval_step(self, args, eval_dataset, step, model, inputs, outputs):
    def on_eval_step(self, args, model, step, batch, batch_features):
        #  outputs = self.on_train_step(args, model, step, batch)

        inputs = self.batch_to_inputs(args, batch)
        attention_mask = inputs['attention_mask']

        outputs = model(**inputs)
        loss, logits = outputs[:2]

        # --------------------------------------
        # NCRFPP CRF
        if args.crf_type == 'ncrfpp':
            if args.n_gpu > 1:
                scores, tags = model.module.crf._viterbi_decode(
                    logits, inputs["attention_mask"])
            else:
                scores, tags = model.crf._viterbi_decode(
                    logits, inputs["attention_mask"])
            tags = tags.cpu().numpy().tolist()
        elif args.crf_type == 'pytorch-crf':
            tags = model.crf.decode(logits)
            #  tags = model.crf.decode(logits.transpose(0, 1))
        elif args.crf_type == 'old_crf':
            tags, _ = model.crf.obtain_labels(logits, inputs['input_lens'])
        elif args.crf_type == 'new_crf':
            tags = model.crf.decode(logits, inputs['attention_mask'])
            tags = tags.cpu().numpy().tolist()[0]
        elif args.crf_type == 'lstm_crf':
            tags = model.crf(logits)

        # --------------------------------------
        #  seq_path = model.crf.decode(logits, attention_mask)

        # --------------------------------------
        # No CRF
        #  tags = np.argmax(logits.cpu(), axis=2)
        #  tags = tags.numpy().tolist()

        # --------------------------------------
        # pytorch-crf
        #  tags = model.crf.decode(logits)
        #  tags = model.crf.decode(logits.transpose(0, 1))

        # --------------------------------------
        # local CRF
        #  tags, _ = model.crf_0.obtain_labels(logits, inputs['input_lens'])

        # --------------------------------------
        # NCRFPP CRF

        #  if args.n_gpu > 1:
        #      scores, tags = model.module.crf._viterbi_decode(
        #          logits, inputs["attention_mask"])
        #  else:
        #      scores, tags = model.crf._viterbi_decode(logits,
        #                                               inputs["attention_mask"])
        #  tags = tags.cpu().numpy().tolist()

        # --------------------------------------
        # local new CRF
        #  tags = model.crf.decode(logits, inputs['attention_mask'])
        #  tags = tags.cpu().numpy().tolist()[0]

        # --------------------------------------
        # torchcrf
        #  tags = model.crf.decode(logits)  #, inputs['attention_mask'].bool())

        # --------------------------------------
        # TorchCRF
        #  tags = model.crf.viterbi_decode(logits,
        #                                  inputs['attention_mask'].bool())

        # --------------------------------------
        # CRFSLTK
        #  _, tags = model.crf(logits, inputs['attention_mask'].bool())

        # --------------------------------------
        # lstm_crf
        #  tags = model.crf(logits)

        #  #  logger.debug(f"tags_0: {tags_0}")
        #  #  logger.info(f"tags_0 size: {np.array(tags_0).shape}")
        #
        #  # logits: [batch_size, seq_len, num_labels]
        #  logger.info(f"logits.shape: {logits.shape}")
        #
        #  #  logits = logits.transpose(0, 1)

        #  # tags: [batch_size, seq_len]
        #
        #  #  tags = torch.LongTensor(tags).transpose(0, 1).cpu().numpy().tolist()
        #  tags = torch.LongTensor(tags).cpu().numpy().tolist()
        #

        #  #  logger.debug(f"tags: {tags}")
        #  #  logger.info(f"tags size: {np.array(tags).shape}")

        # 评估预测结果
        out_label_ids = inputs['labels'].cpu().numpy().tolist()

        #  logger.debug(f"out_label_ids: {out_label_ids}")
        #  logger.debug(f"tags: {tags}")
        for i, label in enumerate(out_label_ids):
            temp_1 = []
            temp_2 = []
            for j, m in enumerate(label):
                if j == 0:
                    continue
                elif out_label_ids[i][j] == args.label2id['[SEP]']:
                    #  logger.info(f"temp_2: {temp_2}")
                    self.metric.update(pred_paths=[temp_2],
                                       label_paths=[temp_1])
                    break
                else:
                    temp_1.append(args.id2label[out_label_ids[i][j]])
                    temp_2.append(args.id2label[tags[i][j]])
                    #  logger.info(f"temp_2: {temp_2}")

        eval_info, _ = self.metric.result()
        results = {f'{key}': value for key, value in eval_info.items()}

        #  batch_preds = tags
        #  for i in range(len(batch_preds)):
        #      preds = batch_preds[i]
        #      R = preds_to_entities(preds, id2label=args.id2label)
        #      T = batch_features[i].subjects
        #      if R != T:
        #          self.error_entities.append((R, T))

        return (loss, ), results

    def on_predict_start(self, args, test_dataset):
        self.pred_results = []

    #  def on_predict_step(self, args, test_dataset, step, model, inputs,
    #                      outputs):
    def on_predict_step(self, args, model, step, batch):
        inputs = self.batch_to_inputs(args, batch, known_labels=False)
        attention_mask = inputs['attention_mask']

        outputs = model(**inputs)
        #  logits = outputs[0][1]
        logits = outputs[0]

        # --------------------------------------
        # NCRFPP CRF
        if args.crf_type == 'ncrfpp':
            _, batch_preds = model.crf._viterbi_decode(
                logits, inputs["attention_mask"])
            batch_preds = batch_preds.cpu().numpy().tolist()
        elif args.crf_type == 'pytorch-crf':
            batch_preds = model.crf.decode(
                logits)  #, inputs['attention_mask'].bool())
        elif args.crf_type == 'old_crf':
            batch_preds, _ = model.crf_0.obtain_labels(logits,
                                                       inputs['input_lens'])
        elif args.crf_type == 'new_crf':
            batch_preds = model.crf.decode(logits, inputs['attention_mask'])
            batch_preds = batch_preds.cpu().numpy().tolist()[0]
        elif args.crf_type == 'lstm_crf':
            tags = model.crf(logits)

        #  batch_preds = model.crf.decode(logits, attention_mask)

        #  batch_preds, _ = model.crf_0._obtain_labels(logits, args.id2label,
        #                                              inputs['input_lens'])

        # ------ local CRF0
        #  batch_preds, _ = model.crf_0.obtain_labels(logits,
        #                                             inputs['input_lens'])

        # ------ local CRF
        #  batch_preds = model.crf.decode(logits, inputs['attention_mask'])
        #  batch_preds = batch_preds.cpu().numpy().tolist()[0]

        # --------------------------------------
        # torchcrf
        #  batch_preds = model.crf.decode(logits)  #, inputs['attention_mask'].bool())

        # --------------------------------------
        # TorchCRF
        #  batch_preds = model.crf.viterbi_decode(logits,
        #                                   inputs['attention_mask'].bool())

        # --------------------------------------
        # lstm_crf
        #  tags = model.crf(logits)

        # --------------------------------------
        # NCRFPP CRF
        #  _, batch_preds = model.crf._viterbi_decode(logits,
        #                                             inputs["attention_mask"])
        #  batch_preds = batch_preds.cpu().numpy().tolist()

        # --------------------------------------
        # CRFSLTK
        #  _, tags = model.crf(logits, inputs['attention_mask'].bool())

        #  batch_preds = np.argmax(batch_preds, axis=2)

        #  batch_preds = model.crf.decode(logits)

        def to_entities(args, preds):
            #  logger.debug(f"{preds}")

            #  preds = [args.id2label[x] for x in preds]
            #  label_entities = get_entities(preds,
            #                                args.id2label,
            #                                args.markup,
            #                                autofix=args.autofix)
            label_entities = preds_to_entities(preds, id2label=args.id2label)

            #  logger.debug(f"{label_entities}")
            json_d = {}
            json_d['id'] = step
            preds = [args.id2label[x] for x in preds]
            json_d['tag_seq'] = " ".join(preds)
            json_d['entities'] = label_entities

            #  logger.debug(f"{json_d}")

            return json_d

        #  preds = preds[0][1:-1]  # [CLS]XXXX[SEP]
        #  json_d = to_entities(args, preds)
        #  self.pred_results.append(json_d)

        for preds in batch_preds:
            preds = preds[1:-1]  # [CLS]XXXX[SEP]
            json_d = to_entities(args, preds)
            self.pred_results.append(json_d)

    def on_predict_end(self, args, test_dataset):
        return self.pred_results

    def on_eval_end(self, args, eval_dataset):
        from ...utils.ner_utils import get_ner_results
        results = get_ner_results(self.metric)
        return results



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import csv, json
from loguru import logger
from collections import Counter
import torch
from transformers import BertTokenizer


class CNerTokenizer(BertTokenizer):
    def __init__(self, vocab_file, do_lower_case=False):
        super().__init__(vocab_file=str(vocab_file),
                         do_lower_case=do_lower_case)
        self.vocab_file = str(vocab_file)
        self.do_lower_case = do_lower_case

    def tokenize(self, text):
        _tokens = []
        for c in text:
            if self.do_lower_case:
                c = c.lower()
            if c in self.vocab:
                _tokens.append(c)
            else:
                _tokens.append('[UNK]')
        return _tokens


class DataProcessor(object):
    def __init__(self):
        self.lines = []

    """Base class for data converters for sequence classification data sets."""

    def get_train_examples(self, data_dir):
        """Gets a collection of `InputExample`s for the train set."""
        raise NotImplementedError()

    def get_dev_examples(self, data_dir):
        """Gets a collection of `InputExample`s for the dev set."""
        raise NotImplementedError()

    def get_labels(self):
        """Gets the list of labels for this data set."""
        raise NotImplementedError()

    @classmethod
    def _read_tsv(cls, input_file, quotechar=None):
        """Reads a tab separated value file."""
        with open(input_file, "r", encoding="utf-8-sig") as f:
            reader = csv.reader(f, delimiter="\t", quotechar=quotechar)
            lines = []
            for line in reader:
                lines.append(line)
            return lines

    @classmethod
    def _read_json(cls, input_file):
        """Reads a json list file."""
        with open(input_file, "r") as f:
            reader = f.readlines()
            lines = []
            for line in reader:
                lines.append(json.loads(line.strip()))
            return lines

    @classmethod
    def _read_text(self, input_file):
        lines = []
        with open(input_file, 'r') as f:
            words = []
            labels = []
            for line in f:
                if line.startswith("-DOCSTART-") or line == "" or line == "\n":
                    if words:
                        lines.append({"words": words, "labels": labels})
                        words = []
                        labels = []
                else:
                    splits = line.split(" ")
                    words.append(splits[0])
                    if len(splits) > 1:
                        labels.append(splits[-1].replace("\n", ""))
                    else:
                        # Examples could have no label for mode = "test"
                        labels.append("O")
            if words:
                lines.append({"words": words, "labels": labels})

        self.lines = lines
        return lines


#  def get_entity_bios_new(seq, id2label):
#      """Gets entities from sequence.
#      note: BIOS
#      Args:
#          seq (list): sequence of labels.
#      Returns:
#          list: list of (chunk_type, chunk_start, chunk_end).
#      Example:
#          # >>> seq = ['B-PER', 'I-PER', 'O', 'S-LOC']
#          # >>> get_entity_bios(seq)
#          [['PER', 0,1], ['LOC', 3, 3]]
#      """
#      hanging = -1
#      chunks = []
#      chunk = [-1, -1, -1]
#      for indx, tag in enumerate(seq):
#          if not isinstance(tag, str):
#              tag = id2label[tag]
#          if tag.startswith("S-"):
#              if chunk[2] != -1:
#                  chunks.append(chunk)
#              chunk = [-1, -1, -1]
#              chunk[1] = indx
#              chunk[2] = indx
#              chunk[0] = ''.join(tag.split('-')[1:])
#              chunks.append(chunk)
#              chunk = [-1, -1, -1]
#              hanging = -1
#          if tag.startswith("B-"):
#              if chunk[2] != -1:
#                  chunks.append(chunk)
#              chunk = [-1, -1, -1]
#              chunk[1] = indx
#              chunk[0] = '-'.join(tag.split('-')[1:])
#              hangle = -1
#          elif tag.startswith('I-') and chunk[1] != -1:
#              _type = '-'.join(tag.split('-')[1:])
#              #  if _type == chunk[0]:
#              chunk[2] = indx
#              if indx == len(seq) - 1:
#                  chunks.append(chunk)
#              hanging = 0
#          elif tag.startswith('I-') and chunk[1] == -1 and indx > 0:
#              _type = '-'.join(tag.split('-')[1:])
#              chunk[0] = _type
#              chunk[1] = indx - 1
#              chunk[2] = indx
#              hanging = 0
#          else:
#              if hanging >= 0:
#                  hanging += 1
#                  if hanging >= 3:
#                      if chunk[2] != -1:
#                          chunks.append(chunk)
#                      chunk = [-1, -1, -1]
#                      hanging = -1
#
#          #  else:
#          #      if chunk[2] != -1:
#          #          chunks.append(chunk)
#          #      chunk = [-1, -1, -1]
#      if chunk[2] != -1:
#          chunks.append(chunk)
#      return chunks


def fix_seq_IOOI(seq, id2label):
    # fix ["I-xxx", 'O', "I-xxx"] to ['I-xxx', 'I-xxx', 'I-xxx']
    if len(seq) < 4:
        return seq

    for i in range(0, len(seq) - 4):
        tag0, tag1, tag2 = seq[i:i + 4]
        #  if not isinstance(tag0, str):
        #      tag0 = id2label[tag0]
        #      tag1 = id2label[tag1]
        #      tag2 = id2label[tag2]
        if tag1 == 'O' and tag0.startswith('I-') and tag1.startswith('I-'):
            seq[i + 1] = tag0
            seq[i + 2] = tag0
            i += 3
    return seq


def fix_seq_BOI(seq, id2label):
    # fix ["I-xxx", 'O', "I-xxx"] to ['I-xxx', 'I-xxx', 'I-xxx']
    if len(seq) < 3:
        return seq

    for i in range(0, len(seq) - 3):
        tag0, tag1, tag2 = seq[i:i + 3]
        #  if not isinstance(tag0, str):
        #      tag0 = id2label[tag0]
        #      tag1 = id2label[tag1]
        #      tag2 = id2label[tag2]
        if tag1 == 'O' and tag0.startswith('B-') and tag2.startswith('I-'):
            seq[i + 1] = 'I-' + tag0[2:]
            i += 2
    return seq


def fix_seq_IOI(seq, id2label):
    # fix ["I-xxx", 'O', "I-xxx"] to ['I-xxx', 'I-xxx', 'I-xxx']
    if len(seq) < 3:
        return seq

    for i in range(0, len(seq) - 3):
        tag0, tag1, tag2 = seq[i:i + 3]
        #  if not isinstance(tag0, str):
        #      tag0 = id2label[tag0]
        #      tag1 = id2label[tag1]
        #      tag2 = id2label[tag2]
        if tag1 == 'O' and tag0.startswith('I-') and tag2.startswith('I-'):
            seq[i + 1] = tag0
            i += 2
    return seq


def fix_seq_OI(seq, id2label):
    # fix ["I-xxx", 'O', "I-xxx"] to ['I-xxx', 'I-xxx', 'I-xxx']
    if len(seq) < 2:
        return seq

    for i in range(0, len(seq) - 2):
        tag0, tag1 = seq[i:i + 2]
        #  if not isinstance(tag0, str):
        #      tag0 = id2label[tag0]
        #      tag1 = id2label[tag1]
        if tag0 == 'O' and tag1.startswith('I-'):
            seq[i] = 'B-' + tag1[2:]
            i += 2
    return seq


def fix_seq_OII(seq, id2label):
    # fix ["I-xxx", 'O', "I-xxx"] to ['I-xxx', 'I-xxx', 'I-xxx']
    if len(seq) < 3:
        return seq

    for i in range(0, len(seq) - 3):
        tag0, tag1, tag2 = seq[i:i + 3]
        #  if not isinstance(tag0, str):
        #      tag0 = id2label[tag0]
        #      tag1 = id2label[tag1]
        #      tag2 = id2label[tag2]
        if tag0 == 'O' and tag1.startswith('I-') and tag2.startswith('I-'):
            seq[i] = 'B-' + tag1[2:]
            i += 3
    return seq


def fix_seq_XOI(seq, id2label, num_O=1):
    if len(seq) < num_O + 2:
        return seq

    #  logger.debug(f"{seq}")
    for i in range(0, len(seq) - num_O - 1):
        tags = [x for x in seq[i:i + num_O + 2]]
        #  if not isinstance(tags[0], str):
        #      tags = [id2label[x] for x in tags]
        tags = [x[:1] for x in tags]
        if tags[0] != 'O' and tags[0] != 'S' and tags[-1] == 'I':
            if tags[1:num_O + 1] == ['O'] * num_O:
                for j in range(1, num_O + 1):
                    seq[i + j] = 'I-' + seq[i][2:]
    return seq


def get_entity_bios(seq, id2label, autofix=False):
    """Gets entities from sequence.
    note: BIOS
    Args:
        seq (list): sequence of labels.
    Returns:
        list: list of (chunk_type, chunk_start, chunk_end).
    Example:
        # >>> seq = ['B-PER', 'I-PER', 'O', 'S-LOC']
        # >>> get_entity_bios(seq)
        [['PER', 0,1], ['LOC', 3, 3]]
    """

    #  logger.debug(f"{seq}")
    #  seq = [id2label[x] for x in seq]
    if autofix:
        seq = fix_seq_XOI(seq, id2label, num_O=3)
        seq = fix_seq_XOI(seq, id2label, num_O=2)
        seq = fix_seq_XOI(seq, id2label, num_O=1)
        #  #  seq = fix_seq_BOI(seq, id2label)
        #  #  seq = fix_seq_IOI(seq, id2label)
        #  #  seq = fix_seq_OI(seq, id2label)
        seq = fix_seq_OII(seq, id2label)

    chunks = []
    chunk = [-1, -1, -1]
    for indx, tag in enumerate(seq):
        #  if not isinstance(tag, str):
        #  tag = id2label[tag]
        if tag.startswith("S-"):
            if chunk[2] != -1:
                chunks.append(chunk)
            chunk = [-1, -1, -1]
            chunk[1] = indx
            chunk[2] = indx
            chunk[0] = '-'.join(tag.split('-')[1:])
            chunks.append(chunk)
            chunk = (-1, -1, -1)
        if tag.startswith("B-"):
            if chunk[2] != -1:
                chunks.append(chunk)
            chunk = [-1, -1, -1]
            chunk[1] = indx
            chunk[0] = '-'.join(tag.split('-')[1:])
        elif tag.startswith('I-') and chunk[1] != -1:
            _type = '-'.join(tag.split('-')[1:])
            #  if _type == chunk[0]:
            #      chunk[2] = indx
            chunk[2] = indx
            if indx == len(seq) - 1:
                chunks.append(chunk)
        else:
            if chunk[2] != -1:
                chunks.append(chunk)
            chunk = [-1, -1, -1]
    return chunks


def get_entity_bio(seq, id2label):
    """Gets entities from sequence.
    note: BIO
    Args:
        seq (list): sequence of labels.
    Returns:
        list: list of (chunk_type, chunk_start, chunk_end).
    Example:
        seq = ['B-PER', 'I-PER', 'O', 'B-LOC']
        get_entity_bio(seq)
        #output
        [['PER', 0,1], ['LOC', 3, 3]]
    """
    chunks = []
    chunk = [-1, -1, -1]
    for indx, tag in enumerate(seq):
        if not isinstance(tag, str):
            tag = id2label[tag]
        if tag.startswith("B-"):
            if chunk[2] != -1:
                chunks.append(chunk)
            chunk = [-1, -1, -1]
            chunk[1] = indx
            chunk[0] = tag.split('-')[1]
            chunk[2] = indx
            if indx == len(seq) - 1:
                chunks.append(chunk)
        elif tag.startswith('I-') and chunk[1] != -1:
            _type = tag.split('-')[1]
            if _type == chunk[0]:
                chunk[2] = indx

            if indx == len(seq) - 1:
                chunks.append(chunk)
        else:
            if chunk[2] != -1:
                chunks.append(chunk)
            chunk = [-1, -1, -1]
    return chunks


def get_entities(seq, id2label, markup='bios', autofix=False):
    '''
    :param seq:
    :param id2label:
    :param markup:
    :return:
    '''
    assert markup in ['bio', 'bios']
    if markup == 'bio':
        return get_entity_bio(seq, id2label)
    else:
        return get_entity_bios(seq, id2label, autofix)


def bert_extract_item(start_logits, end_logits):
    S = []
    start_pred = torch.argmax(start_logits, -1).cpu().numpy()[0][1:-1]
    end_pred = torch.argmax(end_logits, -1).cpu().numpy()[0][1:-1]
    for i, s_l in enumerate(start_pred):
        if s_l == 0:
            continue
        for j, e_l in enumerate(end_pred[i:]):
            if s_l == e_l:
                S.append((s_l, i, i + j))
                break
    return S


class SeqEntityScore(object):
    def __init__(self, id2label, markup='bios', autofix=False):
        self.id2label = id2label
        self.markup = markup
        self.autofix = autofix
        self.reset()

    def reset(self):
        self.origins = []
        self.founds = []
        self.rights = []

    def compute(self, origin, found, right):
        recall = 0 if origin == 0 else (right / origin)
        precision = 0 if found == 0 else (right / found)
        f1 = 0. if recall + precision == 0 else (2 * precision *
                                                 recall) / (precision + recall)
        return recall, precision, f1

    def result(self):
        class_info = {}
        origin_counter = Counter([x[0] for x in self.origins])
        found_counter = Counter([x[0] for x in self.founds])
        right_counter = Counter([x[0] for x in self.rights])
        for type_, count in origin_counter.items():
            origin = count
            found = found_counter.get(type_, 0)
            right = right_counter.get(type_, 0)
            recall, precision, f1 = self.compute(origin, found, right)
            class_info[type_] = {
                "acc": round(precision, 4),
                'recall': round(recall, 4),
                'f1': round(f1, 4)
            }
        origin = len(self.origins)
        found = len(self.founds)
        right = len(self.rights)
        recall, precision, f1 = self.compute(origin, found, right)
        return {'acc': precision, 'recall': recall, 'f1': f1}, class_info

    def update(self, label_paths, pred_paths):
        '''
        labels_paths: [[],[],[],....]
        pred_paths: [[],[],[],.....]

        :param label_paths:
        :param pred_paths:
        :return:
        Example:
            >>> labels_paths = [['O', 'O', 'O', 'B-MISC', 'I-MISC', 'I-MISC', 'O'], ['B-PER', 'I-PER', 'O']]
            >>> pred_paths = [['O', 'O', 'B-MISC', 'I-MISC', 'I-MISC', 'I-MISC', 'O'], ['B-PER', 'I-PER', 'O']]
        '''
        #  logger.debug(f"label_paths: {label_paths}")
        #  logger.debug(f"pred_paths: {pred_paths}")
        for label_path, pre_path in zip(label_paths, pred_paths):
            label_entities = get_entities(label_path, self.id2label,
                                          self.markup, self.autofix)
            pre_entities = get_entities(pre_path, self.id2label, self.markup,
                                        self.autofix)
            self.origins.extend(label_entities)
            self.founds.extend(pre_entities)
            self.rights.extend([
                pre_entity for pre_entity in pre_entities
                if pre_entity in label_entities
            ])



# -*- coding: utf-8 -*-
import os
import re, pickle
import pandas as pd
import numpy as np
from loguru import logger


def determine_sep(filename):
    sep = ','
    idx0 = filename.rfind('.')
    if idx0 > 0:
        if filename[idx0 + 1:] == 'tsv':
            sep = '\t'
            logger.debug(f"Determine sep is '\\t', filename: {filename}")
    return sep


def load_train_data(raw_train_data_file: str,
                    reverse_text: bool = False) -> pd.DataFrame:
    sep = determine_sep(raw_train_data_file)

    df_train = pd.read_csv(raw_train_data_file,
                           names=['ID', 'text', 'category', 'subject'],
                           sep=sep,
                           header=None)
    print(f"df_train.shape: {df_train.shape}")
    df_train = df_train[df_train.category != "其他"]

    #  df_train.text = df_train.text.map(
    #      lambda x: re.sub("[^\w\*\-\+\.\(\)]", " ", x))
    #  df_train.text = df_train.text.map(
    #      lambda x: re.sub("[^\w\*\-\+\.\(\)\（\）]", " ", x))

    if reverse_text:
        df_train.text = df_train.text.map(lambda x: x[::-1])
        df_train.subject = df_train.subject.map(lambda x: x[::-1]
                                                if type(x) == str else "NaN")

    return df_train


def get_train_data_list(df_data: pd.DataFrame) -> list:
    return [(t, c, s, i) for t, c, s, i in zip(df_data.text, df_data.category,
                                               df_data.subject, df_data.ID)]


def load_test_data(raw_test_data_file: str,
                   reverse_text: bool = False) -> pd.DataFrame:
    sep = determine_sep(raw_test_data_file)
    df_test = pd.read_csv(raw_test_data_file,
                          names=['ID', 'text', 'category', 'subject'],
                          sep=sep,
                          header=None)
    df_test.text = df_test.text.fillna("")
    #  df_test.text = df_test.text.map(
    #      lambda x: re.sub("[^\w\*\-\+\.\(\)\（\）]", " ", x))
    if reverse_text:
        df_test.text = df_test.text.map(lambda x: x[::-1])

    return df_test


def get_test_data_list(df_test):
    return [(i, t, c)
            for i, t, c in zip(df_test.ID, df_test.text, df_test.category)]


def save_cross_validation_config(df_train, n_splits, filename):
    random_order = [x for x in range(df_train.shape[0])]
    #  np.random.shuffle(random_order)
    random_order = np.random.permutation(random_order)

    from sklearn import preprocessing
    label_encoder = preprocessing.LabelEncoder()
    print(f"{df_train['category'].unique()}")
    y = label_encoder.fit_transform(df_train['category'])

    def cross_validation(df_train, y, n_splits=5):
        from sklearn.model_selection import StratifiedKFold
        cv_ids = []
        kf = StratifiedKFold(n_splits=n_splits)
        for train, test in kf.split(df_train, y):
            cv_ids.append((train, test))
        return cv_ids

    cv_ids = cross_validation(df_train, y, n_splits=n_splits)

    # Save cross validation ids

    with open(filename, 'wb') as F:
        pickle.dump((random_order, cv_ids), F)

    return random_order, cv_ids


def load_cross_validation_config(filename):
    with open(filename, 'rb') as F:
        random_order, cv_ids = pickle.load(F)
    return random_order, cv_ids


def get_cross_validation_config(df_train, total_folds, train_folds_file):
    #  if os.path.isfile(train_folds_file):
    #      random_order, cv_ids = load_cross_validation_config(
    #          filename=train_folds_file)
    #  else:
    random_order, cv_ids = save_cross_validation_config(
        df_train, n_splits=total_folds, filename=train_folds_file)
    return (random_order, cv_ids)


def get_train_dev_data(df_train, selected_fold, random_order, cv_ids):
    train_ids, test_ids = cv_ids[selected_fold]

    df_train_random = df_train.iloc[random_order, :]

    df_train_data = df_train_random.iloc[train_ids, :]
    df_dev_data = df_train_random.iloc[test_ids, :]

    train_data = get_train_data_list(df_train_data)
    dev_data = get_train_data_list(df_dev_data)

    # 交换训练集与验证集，用小的数据集训练，去预测大的数据集

    #  dev_order = [x for x in range(df_dev_data.shape[0])]
    #  np.random.shuffle(dev_order)
    #  df_dev_random = df_train.iloc[dev_order, :]
    #
    #  dev_data = get_train_data_list(df_dev_random)
    #  train_data = get_train_data_list(df_dev_data)

    #  train_data = [
    #      (t, c, s, i)
    #      for t, c, s, i in zip(df_train_data.text, df_train_data.category,
    #                            df_train_data.subject, df_train_data.ID)
    #  ]
    #  dev_data = [(t, c, s, i)
    #              for t, c, s, i in zip(df_dev_data.text, df_dev_data.category,
    #                                    df_dev_data.subject, df_dev_data.ID)]
    return train_data, dev_data



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import torch
from torch import nn
import torch.nn.functional as F


def mish(input):
    '''
    Applies the mish function element-wise:
    mish(x) = x * tanh(softplus(x)) = x * tanh(ln(1 + exp(x)))
    See additional documentation for mish class.
    '''
    return input * torch.tanh(F.softplus(input))


class Mish(nn.Module):
    '''
    Applies the mish function element-wise:
    mish(x) = x * tanh(softplus(x)) = x * tanh(ln(1 + exp(x)))
    Shape:
        - Input: (N, *) where * means, any number of additional
          dimensions
        - Output: (N, *), same shape as the input
    Examples:
        >>> m = Mish()
        >>> input = torch.randn(2)
        >>> output = m(input)
    '''

    def __init__(self):
        '''
        Init method.
        '''
        super().__init__()

    def forward(self, input):
        '''
        Forward pass of the function.
        '''
        return mish(input)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

#%% ------------------------------------------------------------
import os, sys
import re, json, pickle
import numpy as np
import pandas as pd
import warnings
warnings.filterwarnings('ignore')
#%matplotlib inline
#  sns.set(style='white', context='notebook', palette='deep')

import tensorflow as tf
import keras.backend as K
from keras.losses import categorical_crossentropy
from keras.callbacks import Callback, EarlyStopping

from loguru import logger
from tqdm import tqdm

from tf_utils import get_gpu_info
#  from data import load_train_data
#  from data import load_cross_validation_config, save_cross_validation_config
#  from data import get_cross_validation_config, get_train_dev_data

from .seqlabel_model import BertTokenizer, BertModel
from .predict import extract_labels
from .seqlabel_utils import ensure_empty_directory, list_find, longest_common_substring


#%% ------------------------------------------------------------
class TrainData():
    def __init__(self, data, tokenizer, max_seq_length, batch_size=32):
        self.data = data
        self.tokenizer = tokenizer
        self.batch_size = batch_size
        self.steps = len(self.data) // self.batch_size
        if len(self.data) % self.batch_size != 0:
            self.steps += 1
        self.tag_tokens = self._tokenize(self.data, self.tokenizer,
                                         max_seq_length)
        self.max_seq_length = max_seq_length

    def _tokenize(self, data, tokenizer, max_seq_length):
        tag_tokens = []
        for i, d in enumerate(data):
            #  text, c = d[0][:max_seq_length], d[1]
            text, c = d[0], d[1]
            #  text += " " * (max_seq_length - len(text))
            text = u'___%s___%s' % (c, text)
            text = text[:max_seq_length]

            tokens = tokenizer.tokenize(text)
            num_tokens = len(tokens)

            e = d[2]
            e_tokens = tokenizer.tokenize(e)[1:-1]
            s1, s2 = np.zeros(num_tokens), np.zeros(num_tokens)
            start = list_find(tokens, e_tokens)
            if start != -1:
                end = start + len(e_tokens) - 1
                #  print(f"start:{start}, end:{end}")
                s1[start] = 1
                s2[end] = 1
            tag_tokens.append((text, s1.tolist(), s2.tolist()))
        return tag_tokens

    def __len__(self):
        return self.steps

    def __iter__(self, max_seq_length):
        def seq_padding(X, padding=0):
            L = [len(x) for x in X]
            ML = max(L)
            return np.array([
                np.concatenate([x, [padding] *
                                (ML - len(x))]) if len(x) < ML else x
                for x in X
            ])

        while True:
            idxs = [x for x in range(len(self.data))]
            np.random.shuffle(idxs)
            X1, X2, S1, S2 = [], [], [], []
            for i in idxs:
                text, s1, s2 = self.tag_tokens[i]
                if not (s1 is None or s2 is None):
                    x1, x2 = self.tokenizer.encode(first=text)
                    X1.append(x1)
                    X2.append(x2)
                    S1.append(np.array(s1))
                    S2.append(np.array(s2))
                    if len(X1) == self.batch_size or i == idxs[-1]:
                        #  print(f"X1, X2, S1, S2")
                        X1 = seq_padding(X1)
                        X2 = seq_padding(X2)
                        S1 = seq_padding(S1)
                        S2 = seq_padding(S2)
                        yield [X1, X2, S1, S2], None
                        X1, X2, S1, S2 = [], [], [], []


#%% ------------------------------------------------------------
class Evaluate(Callback):
    def __init__(self, args, bert_dev_data, extract_entity, bert_model,
                 learning_rate, min_learning_rate, strictly_equal):
        self.args = args
        self.bert_dev_data = bert_dev_data
        self.extract_entity = extract_entity
        self.bert_model = bert_model
        self.learning_rate = learning_rate
        self.min_learning_rate = min_learning_rate
        self.strictly_equal = strictly_equal
        #  self.train_ACC = []
        self.dev_ACC = []
        #  self.train_best = 0.
        self.best_loss = float('inf')
        self.last_loss = float('inf')
        self.best_f1 = 0.0
        self.last_f1 = 0.0
        self.num_loss_up = 0
        self.passed = 0

    def on_batch_begin(self, batch, logs=None):

        if 'COLAB_TPU_ADDR' in os.environ:
            self.on_epoch_begin_tpu(batch, logs)
            return
        """第一个epoch用来warmup，第二个epoch把学习率降到最低
        """
        if self.passed < self.params['steps']:
            lr = (self.passed + 1.) / self.params['steps'] * self.learning_rate
            K.set_value(self.model.optimizer.lr, lr)
            self.passed += 1
        elif self.params['steps'] <= self.passed < self.params['steps'] * 2:
            lr = (2 - (self.passed + 1.) / self.params['steps']) * (
                self.learning_rate - self.min_learning_rate)
            lr += self.min_learning_rate
            K.set_value(self.model.optimizer.lr, lr)
            self.passed += 1
        #  print(
        #      f"on_batch_begin() learning rate: {lr:.6f}, passed: {self.passed}, params['steps']: {self.params['steps']}"
        #  )

    def get_dev_loss(self, predict_tokens):
        # ------- loss -------
        def categorical_crossentropy_loss(y_true: np.array,
                                          y_pred: np.array) -> float:
            """
            计算分类交叉熵（适用于多分类，张量版本）
            y_true: 训练标签，[[1,0,0], [0,1,0],...]。
            y_pred: 预测标签，[[1,0,0], [0,1,0],...]。
            返回交叉熵张量，形如[0.983423]。
            """

            #  # 传入单个标签时，转换成标签列表的形式。
            #  if len(y_true.shape) == 1:
            #      y_true = y_true.reshape((1, y_true.shape[0]))
            #      y_pred = y_pred.reshape((1, y_pred.shape[0]))
            # 生成形如[[0,1,0,...], [0,0,1,...],...]的张量。
            y_true = tf.convert_to_tensor(y_true, dtype='float32')
            y_pred = tf.convert_to_tensor(y_pred, dtype='float32')

            # 根据标签位置距离，计算分类权重。
            i_true = K.argmax(y_true, axis=1)
            i_pred = K.argmax(y_pred, axis=1)
            distance = K.abs(i_true - i_pred)
            length = K.int_shape(y_pred)[1] - 1
            weights = K.cast(distance / length, dtype='float32')

            losses = ((1.0 + weights) *
                      categorical_crossentropy(y_true, y_pred))
            loss = K.eval(K.mean(losses))
            return loss

        subject_tokens = self.bert_dev_data.tag_tokens

        max_s1 = max(len(x) for _, x, _ in subject_tokens)
        max_s2 = max(len(x) for _, _, x in subject_tokens)
        max_k1 = max(len(x) for x, _ in predict_tokens)
        max_k2 = max(len(x) for _, x in predict_tokens)
        max_tokens_len = max([max_s1, max_s2, max_k1, max_k2])

        s1 = []
        s2 = []
        for _, _s1, _s2 in subject_tokens:
            s1.append(_s1 + [0.0] * (max_tokens_len - len(_s1)))
            s2.append(_s2 + [0.0] * (max_tokens_len - len(_s2)))

        k1 = []
        k2 = []
        for _k1, _k2 in predict_tokens:
            k1.append(_k1 + [0.0] * (max_tokens_len - len(_k1)))
            k2.append(_k2 + [0.0] * (max_tokens_len - len(_k2)))

        loss1 = categorical_crossentropy_loss(s1, k1)
        loss2 = categorical_crossentropy_loss(s2, k2)
        loss = loss1 + loss2

        return loss

    def on_epoch_end(self, epoch, logs=None):
        train_acc, wrong_train, train_loss = 0.0, [], 0.0  #self.evaluate(train_data)
        print(f"Evaluating dev data...")
        dev_data = self.bert_dev_data.data

        dev_loss, dev_acc, dev_recall, dev_f1, error_list = self.evaluate_by_ID(
            dev_data)
        #  dev_acc, wrong_eval, dev_loss = self.evaluate(dev_data)

        #  self.train_ACC.append(train_acc)
        self.dev_ACC.append(dev_acc)

        logger.info(f"Epoch {epoch+1}: "
                    f"dev_loss: {dev_loss:.4f}, "
                    f"dev_acc: {dev_acc:.4f}, "
                    f"dev_recall: {dev_recall:.4f}, "
                    f"dev_f1: {dev_f1:.4f}")

        # 保存loss最小的模型
        #  if dev_loss < self.best_loss:
        #      self.best_loss = dev_loss
        # 保存f1最大的模型
        if dev_f1 > self.best_f1:
            self.best_f1 = dev_f1

            best_model_weights_file = get_best_model_weights_file_name(
                selected_fold, self.args.category)
            self.bert_model.train_model.save_weights(best_model_weights_file)
            logger.success(
                f"Saved best model weights to {best_model_weights_file}.")
            #  os.symlink()

            #  wrong_eval_file = get_wrong_eval_file_name(selected_fold)
            #  with open(wrong_eval_file, 'w') as F:
            #      last_ID = ""
            #      for ID, true_text, pred_text in error_list:
            #          if ID != last_ID:
            #              if last_ID:
            #                  F.write("\n")
            #              last_ID = ID
            #          F.write(f"{ID} | \t {true_text} | \t {pred_text}\n")

        #  if dev_loss > self.last_loss:
        #      self.num_loss_up += 1
        #      logger.warning(f"Dev loss keep raise {self.num_loss_up} times.")
        #  else:
        #      self.num_loss_up = 0
        if dev_f1 < self.last_f1:
            self.num_loss_up += 1
            logger.warning(f"Dev f1 keep decreased {self.num_loss_up} times.")
        else:
            self.num_loss_up = 0

        self.last_loss = dev_loss
        self.last_f1 = dev_f1
        #  print("")
    def evaluate_by_ID(self, data_list):
        dict_val_labels = {}
        test_data = []
        for text_in, c_in, label, ID in tqdm(iter(data_list)):
            if ID not in dict_val_labels:
                dict_val_labels[ID] = []
            dict_val_labels[ID].append(label)
            test_data.append((ID, text_in, c_in))

        from predict import extract_labels_from_test_data
        results = extract_labels_from_test_data(
            test_data,
            category=self.args.category,
            top_k=self.args.top_k,
            s_threshold=self.args.s_threshold,
            e_threshold=self.args.e_threshold,
            min_label_len=self.args.min_label_len,
            seg_len=self.args.seg_len,
            seg_backoff=self.args.seg_backoff,
            bert_model=self.bert_model)

        #  print(results)
        dict_labels = {
            ID: [x[0] for x in items]
            for ID, items in results.items()
        }

        def text_similarity(s_true, s_pred, strictly_equal=True):
            if s_true == s_pred:
                return 1.0
            else:
                if strictly_equal:
                    return 0.0
                else:
                    if len(s_true) == 0:
                        return 0.0
                    s1_words = [w for w in s_true]
                    s2_words = [w for w in s_pred]
                    s1_end, s2_end, longest, substring = longest_common_substring(
                        s1_words, s2_words)
                    return longest / max(len(s1_words), len(s2_words))

        error_list = []
        loss = 0.0
        TP = FP = TN = FN = 0.0
        for ID, list_true_labels in dict_val_labels.items():
            tp = fp = tn = fn = 0.0
            list_pred_labels = dict_labels.get(ID, [])
            if list_pred_labels:

                # 严格按顺序对位比较
                #  for i, t_label in enumerate(list_true_labels):
                #      if i <= len(list_pred_labels) - 1:
                #          similarity_score = text_similarity(
                #              t_label,
                #              list_pred_labels[i],
                #              strictly_equal=self.strictly_equal)
                #          assert similarity_score >= 0.0 and similarity_score <= 1.0
                #          if similarity_score < 1.0:
                #              error_list.append(
                #                  (ID, t_label, list_pred_labels[i]))
                #
                #          tp += similarity_score
                #          fp += 1.0 - similarity_score
                #          loss += 1.0 - similarity_score
                #      else:
                #          fp += 1.0
                #          loss += 1.0
                #          error_list.append((ID, t_label, ""))
                #  if len(list_pred_labels) > len(list_true_labels):
                #      loss += (len(list_pred_labels) -
                #               len(list_true_labels)) * 1.0
                #      for pred_label in list_pred_labels[len(list_true_labels):]:
                #          error_list.append((ID, "", pred_label))

                # 取实际标注在预测标注的最大相似度
                for t_label in list_true_labels[:min(len(list_true_labels),
                                                     len(list_pred_labels))]:
                    max_similarity_score = 0.0
                    for p_label in list_pred_labels:
                        similarity_score = text_similarity(
                            t_label,
                            p_label,
                            strictly_equal=self.strictly_equal)
                        if similarity_score > max_similarity_score:
                            max_similarity_score = similarity_score
                    tp += max_similarity_score
                    fp += 1.0 - max_similarity_score
                    loss += 1.0 - max_similarity_score
                if len(list_pred_labels) > len(list_true_labels):
                    loss += (len(list_pred_labels) -
                             len(list_true_labels)) * 1.0
                    for pred_label in list_pred_labels[len(list_true_labels):]:
                        error_list.append((ID, "", pred_label))
                elif len(list_pred_labels) < len(list_true_labels):
                    loss += (len(list_true_labels) -
                             len(list_pred_labels)) * 1.0
                    fp += len(list_true_labels) - len(list_pred_labels)
                    for true_label in list_true_labels[len(list_pred_labels):]:
                        error_list.append((ID, true_label, ""))

            else:
                loss += len(list_true_labels) * 1.0
                fp += len(list_true_labels) * 1.0

            TP += tp
            FP += fp
            TN += tn
            FN += fn
        logger.debug(f"TP: {TP}, FP: {FP}, TN: {TN}, FN: {FN}")
        true_chars = sum(
            [len(labels) for ID, labels in dict_val_labels.items()])
        pred_chars = sum([len(labels) for ID, labels in dict_labels.items()])
        logger.debug(f"true_chars: {true_chars}, " f"pred_chars: {pred_chars}")
        #  if self.strictly_equal:
        #      assert true_chars == TP + FP

        acc = (TP + FN) / pred_chars
        recall = (TP + FN) / true_chars
        f1 = 2 * acc * recall / (acc + recall)

        return loss, acc, recall, f1, error_list

    def evaluate(self, data_list):
        A = 1e-10
        wrong_list = []
        i = 0
        tag_tokens = []
        for d in tqdm(iter(data_list)):
            R, tag_token = self.extract_entity(d[0], d[1])
            tag_tokens.append(tag_token)
            if R == d[2]:
                A += 1
            else:
                wrong_list.append((i, d[2], R))
            i += 1
        dev_loss = self.get_dev_loss(tag_tokens)
        return A / len(data_list), wrong_list, dev_loss


#%% ------------------------------------------------------------
class SeqlabelTrainer:
    def ___init__(self):
        pass

    def _do_train(self,
                  args,
                  bert_train_data,
                  bert_dev_data,
                  selected_fold,
                  force_run=False):
        print(f"Training fold {selected_fold}...")

        best_model_weights_file = get_best_model_weights_file_name(
            selected_fold, self.args.category)
        if os.path.isfile(best_model_weights_file) or os.path.isfile(
                best_model_weights_file + ".index"):
            logger.info(
                f"The fold {selected_fold} has already been trained. Skip train step."
            )
            return

        from tf_utils import enable_gpu_growth
        enable_gpu_growth()

        train_data = bert_train_data.data
        dev_data = bert_dev_data.data

        additional_chars = set()
        for row in train_data + dev_data:
            additional_chars.update(
                re.findall(u'[^\u4e00-\u9fa5a-zA-Z0-9\*]', row[2]))
            #  "[^a-zA-Z0-9\/\.\,\?\(\)\[\]\{\}\u4e00-\u9fa5]"
        if u'，' in additional_chars:
            additional_chars.remove(u'，')

        bert_model = BertModel(args.config_path, args.onfig.checkpoint_path,
                               args.dict_path, self.args.learning_rate,
                               args.max_seq_length)

        bert_model.build_model()

        def _extract_entity(text_in, c_in, bert_model, classes,
                            additional_chars):
            return bert_model.extract_entity(text_in, c_in, additional_chars)

        def extract_entity_func():
            from functools import partial
            return partial(_extract_entity,
                           bert_model=bert_model,
                           classes=classes,
                           additional_chars=additional_chars)

        extract_entity = extract_entity_func()

        logger.info(f"Start train fold {selected_fold}...")
        evaluator = Evaluate(bert_dev_data, extract_entity, bert_model,
                             args.learning_rate, args.min_learning_rate,
                             args.strictly_equal)
        #  es_monitor = 'val_loss'  # acc, loss, val_loss, val_acc
        #  es_patience = 50  # 当early stop被激活，则经过patience个epoch后停止训练。
        #  es_mode = 'min'  # 'auto', 'min', 'max'
        #  es_verbose = 2
        #  early_stopping = EarlyStopping(monitor=es_monitor,
        #                                 patience=es_patience,
        #                                 mode=es_mode,
        #                                 verbose=es_verbose)

        bert_model.train_model.fit_generator(
            bert_train_data.__iter__(args.max_seq_length),
            steps_per_epoch=len(bert_train_data),
            #  epochs=3 * steps_per_update,
            epochs=args.epochs,
            callbacks=[evaluator])

        bert_model = None
        K.clear_session()

        #  tf.reset_default_graph()

        def train(self, args, train_data, dev_data):
            """
            train_data: ID, text, category, entity
            dev_data: same as train_data
            """

            best_model_weights_file = get_best_model_weights_file_name(
                selected_fold, args.category)
            if os.path.isfile(best_model_weights_file) or os.path.isfile(
                    best_model_weights_file + ".index"):
                logger.info(
                    f"The fold {selected_fold} has already been trained. Skip train step."
                )
                return

            tokenizer = BertTokenizer(args.dict_path)
            bert_train_data = TrainData(train_data,
                                        tokenizer,
                                        args.max_seq_length,
                                        batch_size=batch_size)
            bert_dev_data = TrainData(dev_data,
                                      tokenizer,
                                      args.max_seq_length,
                                      batch_size=batch_size)
            self._do_train(
                bert_train_data,
                bert_dev_data,
            )

    def predict(self, args, resume_texts, category, bert_model):

        # results: [(ID, label_text, idx0, idx1, prob0, prob1), ...]
        results = extract_labels(resume_texts,
                   category,
                   top_k=args.top_k,
                   s_threshold=args.s_threshold,
                   e_threshold=args.e_threshold,
                   min_label_len=args.min_label_len,
                   seg_len=args.seg_len,
                   seg_backoff=args.seg_backoff,
                   bert_model=bert_model)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from pathlib import Path
import json
from tqdm import tqdm
from loguru import logger
import numpy as np
import pandas as pd
import torch
from torch.utils.data import DataLoader, TensorDataset
from torch.utils.data import RandomSampler, SequentialSampler
from ...utils.multiprocesses import barrier_leader_process, barrier_member_processes, is_master_process, is_multi_processes
from ...utils import seg_generator
from ..glue_utils import InputExample


class InputFeatures(object):
    """
    A single set of features of data.

    Args:
        input_ids: Indices of input sequence tokens in the vocabulary.
        attention_mask: Mask to avoid performing attention on padding token indices.
            Mask values selected in ``[0, 1]``:
            Usually  ``1`` for tokens that are NOT MASKED, ``0`` for MASKED (padded) tokens.
        token_type_ids: Segment token indices to indicate first and second portions of the inputs.
        label: Label corresponding to the input
    """
    def __init__(self,
                 input_ids,
                 attention_mask=None,
                 token_type_ids=None,
                 label=None):
        self.input_ids = input_ids
        self.attention_mask = attention_mask
        self.token_type_ids = token_type_ids
        self.label = label

    def __repr__(self):
        return str(self.to_json_string())

    def to_dict(self):
        """Serializes this instance to a Python dictionary."""
        output = copy.deepcopy(self.__dict__)
        return output

    def to_json_string(self):
        """Serializes this instance to a JSON string."""
        return json.dumps(self.to_dict(), indent=2, sort_keys=True) + "\n"


#  def load_examples(args,
#                    data_generator,
#                    examples_file,
#                    seg_len=0,
#                    seg_backoff=0):
#
#      examples = []
#
#      for guid, text_a, text_b, label in data_generator(args,
#                                                        examples_file,
#                                                        seg_len=seg_len,
#                                                        seg_backoff=seg_backoff):
#          #  for (seg_text_a, seg_text_b) in seg_generator((text_a, text_b),
#          #                                                seg_len, seg_backoff):
#          #      #  seg_text = seg_text[0]
#          #      examples.append(
#          #          InputExample(guid=guid,
#          #                       text_a=seg_text_a,
#          #                       text_b=seg_text_b,
#          #                       label=label))
#          examples.append(
#              InputExample(guid=guid, text_a=text_a, text_b=text_b, label=label))
#      logger.info(f"Loaded {len(examples)} examples from {examples_file}.")
#
#      return examples


def encode_examples(examples, label2id, tokenizer, max_seq_length):

    texts_a = [e.text_a for e in examples]
    texts_b = [e.text_b for e in examples]
    if not any(texts_b):
        texts_b = None
    #  texts_b = None
    #  if len(examples) > 0 and examples[0].text_b:
    #      texts_b = [e.text_b for e in examples]

    outputs = tokenizer.batch_encode_plus(texts_a,
                                          text_pair=texts_b,
                                          max_length=max_seq_length,
                                          add_special_tokens=True,
                                          return_tensors=None,
                                          pad_to_max_length=True,
                                          return_attention_mask=True,
                                          return_token_type_ids=True)
    all_input_ids, all_attention_mask, all_token_type_ids = outputs[
        'input_ids'], outputs['attention_mask'], outputs['token_type_ids']

    all_lens = [len(input_ids) for input_ids in all_input_ids]

    #  all_labels = [label2id[e.label] if e.label else 0 for e in examples]

    all_labels = []
    for e in examples:
        if isinstance(e.label, list):
            targets = [0] * len(label2id)
            for x in e.label:
                targets[label2id[x]] = 1
            all_labels.append(targets)
        else:
            if e.label:
                all_labels.append(label2id[e.label])
            else:
                all_labels.append(0)

    all_features = [
        InputFeatures(input_ids=input_ids,
                      attention_mask=attention_mask,
                      token_type_ids=token_type_ids,
                      label=label)
        for input_ids, attention_mask, token_type_ids, label in zip(
            all_input_ids, all_attention_mask, all_token_type_ids, all_labels)
    ]
    return (all_features, {
        'input_ids':
        torch.tensor(all_input_ids, dtype=torch.long),
        'attention_mask':
        torch.tensor(all_attention_mask, dtype=torch.long),
        'token_type_ids':
        torch.tensor(all_token_type_ids, dtype=torch.long),
        'lens':
        torch.tensor(all_lens, dtype=torch.long),
        'labels':
        torch.tensor(all_labels, dtype=torch.long), 
    })


def examples_to_dataset(examples, label2id, tokenizer, max_seq_length):

    all_features, outputs = encode_examples(examples, label2id, tokenizer,
                                            max_seq_length)

    dataset = TensorDataset(
        outputs['input_ids'],
        outputs['attention_mask'],
        outputs['token_type_ids'],
        outputs['lens'],
        outputs['labels'],
    )

    return dataset, all_features


#  def convert_examples_to_features(
#      examples,
#      label2id,
#      tokenizer,
#      max_length,
#      output_mode="classification",
#      pad_on_left=False,
#      pad_token=0,
#      pad_token_segment_id=0,
#      mask_padding_with_zero=True,
#  ):
#      """
#      Loads a data file into a list of ``InputFeatures``
#
#      Args:
#          examples: List of ``InputExamples`` or ``tf.data.Dataset`` containing the examples.
#          tokenizer: Instance of a tokenizer that will tokenize the examples
#          max_length: Maximum example length
#          task: GLUE task
#          glue_labels: List of labels. Can be obtained from the processor using the ``processor.get_labels()`` method
#          output_mode: String indicating the output mode. Either ``regression`` or ``classification``
#          pad_on_left: If set to ``True``, the examples will be padded on the left rather than on the right (default)
#          pad_token: Padding token
#          pad_token_segment_id: The segment ID for the padding token (It is usually 0, but can vary such as for XLNet where it is 4)
#          mask_padding_with_zero: If set to ``True``, the attention mask will be filled by ``1`` for actual values
#              and by ``0`` for padded values. If set to ``False``, inverts it (``1`` for padded values, ``0`` for
#              actual values)
#
#      Returns:
#          If the ``examples`` input is a ``tf.data.Dataset``, will return a ``tf.data.Dataset``
#          containing the task-specific features. If the input is a list of ``InputExamples``, will return
#          a list of task-specific ``InputFeatures`` which can be fed to the model.
#
#      """
#
#      features = []
#      for (ex_index, example) in enumerate(tqdm(examples)):
#          len_examples = len(examples)
#          #  if ex_index % 10000 == 0:
#          #      logger.info("Writing example %d/%d" % (ex_index, len_examples))
#
#          inputs = tokenizer.encode_plus(
#              example.text_a,
#              example.text_b,
#              add_special_tokens=True,
#              max_length=max_length,
#              return_token_type_ids=True,
#          )
#          input_ids, token_type_ids = inputs["input_ids"], inputs[
#              "token_type_ids"]
#
#          # The mask has 1 for real tokens and 0 for padding tokens. Only real
#          # tokens are attended to.
#          attention_mask = [1 if mask_padding_with_zero else 0] * len(input_ids)
#
#          # Zero-pad up to the sequence length.
#          padding_length = max_length - len(input_ids)
#          if pad_on_left:
#              input_ids = ([pad_token] * padding_length) + input_ids
#              attention_mask = ([0 if mask_padding_with_zero else 1] *
#                                padding_length) + attention_mask
#              token_type_ids = ([pad_token_segment_id] *
#                                padding_length) + token_type_ids
#          else:
#              input_ids = input_ids + ([pad_token] * padding_length)
#              attention_mask = attention_mask + (
#                  [0 if mask_padding_with_zero else 1] * padding_length)
#              token_type_ids = token_type_ids + ([pad_token_segment_id] *
#                                                 padding_length)
#
#          assert len(input_ids
#                     ) == max_length, "Error with input length {} vs {}".format(
#                         len(input_ids), max_length)
#          assert len(attention_mask
#                     ) == max_length, "Error with input length {} vs {}".format(
#                         len(attention_mask), max_length)
#          assert len(token_type_ids
#                     ) == max_length, "Error with input length {} vs {}".format(
#                         len(token_type_ids), max_length)
#
#          label = 0
#          if example.label is not None:
#              if output_mode == "classification":
#                  label = label2id[example.label]
#              elif output_mode == "regression":
#                  label = float(example.label)
#              else:
#                  raise KeyError(output_mode)
#
#          if ex_index < 3:
#              logger.info("*** Example ***")
#              logger.info("guid: %s" % (example.guid))
#
#              tokens_a = tokenizer.tokenize(example.text_a)
#              logger.info(f"tokens_a: {' '.join([str(x) for x in tokens_a])}")
#              if example.text_b:
#                  tokens_b = tokenizer.tokenize(example.text_b)
#                  logger.info(
#                      f"tokens_b: {' '.join([str(x) for x in tokens_b])}")
#
#              logger.info("input_ids: %s" % " ".join([str(x)
#                                                      for x in input_ids]))
#              logger.info("attention_mask: %s" %
#                          " ".join([str(x) for x in attention_mask]))
#              logger.info("token_type_ids: %s" %
#                          " ".join([str(x) for x in token_type_ids]))
#              logger.info("label: %s (id = %d)" % (example.label, label))
#
#          features.append(
#              InputFeatures(input_ids=input_ids,
#                            attention_mask=attention_mask,
#                            token_type_ids=token_type_ids,
#                            label=label))
#
#      logger.info(f"Leave {len(features)} features")
#      return features
#
#
#  def features_to_dataset(features):
#      # Convert to Tensors and build dataset
#      all_input_ids = torch.tensor([f.input_ids for f in features],
#                                   dtype=torch.long)
#      all_attention_mask = torch.tensor([f.attention_mask for f in features],
#                                        dtype=torch.long)
#      all_token_type_ids = torch.tensor([f.token_type_ids for f in features],
#                                        dtype=torch.long)
#      #  if output_mode == "regression":
#      all_labels = torch.tensor([f.label for f in features], dtype=torch.long)
#      #  elif output_mode == "regression":
#      #      all_labels = torch.tensor([f.label for f in features],
#      #                                dtype=torch.float)
#
#      dataset = TensorDataset(all_input_ids, all_attention_mask,
#                              all_token_type_ids, all_labels)
#      return dataset
#

#  def examples_to_dataset_old(args,
#                              examples,
#                              tokenizer,
#                              max_seq_length,
#                              evaluate=False,
#                              output_mode="classification",
#                              alias=""):
#      # Make sure only the first process in distributed training process the dataset,
#      # and the others will use the cache
#      if not evaluate:
#          barrier_member_processes(args)
#
#      # Load data features from cache or dataset file
#      if args.data_dir:
#          cached_features_file = Path(
#              args.data_dir
#          ) / f"cached_glue_{args.dataset_name}{'_' + alias if alias else ''}"
#      else:
#          cached_features_file = None
#
#      if args.cache_features and cached_features_file and cached_features_file.exists(
#      ) and not args.overwrite_cache:
#          logger.info(
#              f"Loading features from cached file {cached_features_file}")
#          features = torch.load(cached_features_file)
#      else:
#          logger.info(f"Creating features from dataset file at {args.data_dir}")
#
#          features = convert_examples_to_features(
#              examples,
#              args.label2id,
#              tokenizer,
#              max_length=max_seq_length,
#              output_mode=output_mode,
#              pad_on_left=bool(
#                  args.model_type in ["xlnet"]),  # pad on the left for xlnet
#              pad_token=tokenizer.pad_token_id,
#              pad_token_segment_id=tokenizer.pad_token_type_id,
#          )
#
#          if args.cache_features and cached_features_file and is_master_process(
#                  args):
#              logger.info(
#                  f"Saving features into cached file {cached_features_file}")
#              torch.save(features, cached_features_file)
#
#      logger.info(
#          f"convert {len(examples)} examples to {len(features)} features.")
#      # Make sure only the first process in distributed training process the dataset,
#      # and the others will use the cache
#      if not evaluate:
#          barrier_leader_process(args)
#
#      dataset = features_to_dataset(features)
#
#      return dataset

#  # -------------------- GlueExamples --------------------
#  class GlueExamples:
#      def __init__(self, filename, glue_labels, name):
#          self.filename = filename
#          self.name = name
#          self.glue_labels = glue_labels
#          self.examples = self._create_examples(self._read_txt(filename))
#
#          self._idx = 0
#
#      def __iter__(self):
#          return self
#
#      def __next__(self):
#          if self._idx < len(self.examples):
#              e = self.examples[self._idx]
#              self._idx += 1
#              return e
#          else:
#              raise StopIteration
#
#      def __get_item__(self, idx):
#          return self.examples[idx]
#
#      def __len__(self):
#          return len(self.examples)
#
#
# -------------------- DualSentencesDataset --------------------

#  class DualSentencesDataset:
#      """
#      原生只读训练数据集，用于生成模型训练集、验证集、测试集。
#      缺省文件名train_dataset.csv，必须有header，分隔符缺省为'\t'，
#      字段['sid', text_a', 'text_b', 'label', 'category']。
#      """
#
#      columns = ['sid', 'text_a', 'text_b', 'label', 'category']
#
#      def __init__(self,
#                   samples: [list, np.ndarray] = None,
#                   name: str = "default"):
#          self.name = name
#          self.columns = ['sid', 'text_a', 'text_b', 'label', 'category']
#          self.labels = []
#          self.categories = []
#          self.samples = pd.DataFrame([], columns=self.columns)
#
#          if samples is not None:
#              self.read_data(samples)
#
#      def __len__(self):
#          return self.samples.shape[0]
#
#      def info(self):
#          num_samples = self.samples.shape[0]
#
#          t1 = [len(x.text_a) for x in self.samples.itertuples()]
#          t1_mean = np.mean(t1)
#          t1_std = np.std(t1)
#          t1_min = np.min(t1)
#          t1_max = np.max(t1)
#          t2 = [len(x.text_b) for x in self.samples.itertuples()]
#          t2_mean = np.mean(t2)
#          t2_std = np.std(t2)
#          t2_min = np.min(t2)
#          t2_max = np.max(t2)
#
#          print(f"======== {self.name} ========")
#          print(f"Total {num_samples} samples.")
#          if num_samples == 0:
#              return
#
#          print(f"-------- sid --------")
#          num_sid = len(list(set(self.samples.sid.unique())))
#          print(f"{num_sid} unique sid.")
#          num_nan_sid = np.sum(
#              [len(x.sid) == 0 for x in self.samples.itertuples()])
#          print(f"{num_nan_sid} NaN sid.")
#
#          print(f"-------- text_a --------")
#          print(
#              f"mean: {t1_mean:.2f} std: {t1_std:.2f} min: {t1_min} max: {t1_max}"
#          )
#          print(f"-------- text_b --------")
#          print(
#              f"mean: {t1_mean:.2f} std: {t1_std:.2f} min: {t1_min} max: {t1_max}"
#          )
#          print(f"-------- labels --------")
#          print(f"{self.labels}")
#          if len(self.labels) > 0:
#              for label in self.labels:
#                  label_samples = np.sum(self.samples.label == label)
#                  print(
#                      f"{label}: {label_samples} ({label_samples/num_samples:.2f})"
#                  )
#
#          print(f"-------- categories --------")
#          print(f"{self.categories}")
#          if len(self.categories) > 0:
#              for category in self.categories:
#                  category_samples = np.sum(self.samples.category == category)
#                  print(
#                      f"{category}: {category_samples} ({category_samples/num_samples:.2f})"
#                  )
#
#      def read_csv(self,
#                   csv_file: str,
#                   default_label='-1',
#                   default_category='-1',
#                   sep='\t'):
#          df_samples = pd.read_csv(csv_file, sep=sep, dtype=np.str)
#          if 'label' not in df_samples.columns:
#              df_samples['label'] = [default_label] * df_samples.shape[0]
#          if 'category' not in df_samples.columns:
#              df_samples['category'] = [default_category] * df_samples.shape[0]
#
#          if not self._check_has_all_needed_columns(df_samples):
#              raise TypeError(
#                  f"Columns of samples must contain all columns in {self.columns}"
#              )
#
#          self.samples = df_samples.astype(np.str)
#          self._init_samples()
#
#          return True
#
#      def read_data(self, samples: [list, np.ndarray]):
#          if isinstance(samples, list):
#              samples = np.array(samples)
#          print(samples.shape)
#
#          if samples.shape[1] != len(self.columns):
#              raise TypeError(
#                  f"Columns of samples must contain all columns in {self.columns}"
#              )
#          self.samples = pd.DataFrame(samples, columns=self.columns)
#
#          self._init_samples()
#
#          return True
#
#      def _init_samples(self):
#          self.samples.text_a.fillna("")
#          self.samples.text_b.fillna("")
#          self.samples.label.fillna("")
#          self.samples.category.fillna("")
#
#          self.samples.info()
#
#          self.labels = sorted(list(set(self.samples.label.unique())))
#          self.categories = sorted(list(set(self.samples.category.unique())))
#
#      def to_csv(self,
#                 csv_file: str,
#                 columns: list = [],
#                 sep='\t',
#                 shuffle=False,
#                 random_state=None):
#
#          if len(columns) > 0:
#              samples = self.samples[columns]
#          else:
#              samples = self.samples
#              columns = self.columns
#          samples = samples.to_numpy()
#
#          if shuffle:
#              samples = shuffle_list(samples, random_state=random_state)
#
#          df_samples = pd.DataFrame(samples, columns=columns)
#          df_samples[['sid', 'text_a', 'text_b', 'label',
#                      'category']].to_csv(csv_file, index=False, sep=sep)
#          print(f"Dataset {self.name} saved to {csv_file}.")
#
#      def _check_has_all_needed_columns(self, df_samples: pd.DataFrame):
#          return np.sum([x in df_samples.columns
#                         for x in self.columns]) == len(self.columns)
#
#      def split_label(self,
#                      label: str,
#                      rates: list = [0.5],
#                      random_state=None) -> list:
#          samples = self.samples[self.samples.label == label].to_numpy()
#          samples = shuffle_list(samples, random_state=random_state)
#
#          nos = [int(samples.shape[0] * r) for r in rates]
#          last = samples.shape[0] - np.sum(nos)
#          assert (last > 0)
#          nos += [last]
#          pos = np.add.accumulate(nos)
#
#          samples_list = []
#          p0 = 0
#          for p1 in pos:
#              samples_list.append(samples[p0:p1])
#              p0 = p1
#
#          return samples_list
#
#      def split(self, rates: list = [0.5], random_state=None) -> list:
#          num_rates = len(rates) + 1
#          label_samples = []
#          for label in self.labels:
#              X = self.split_label(label, rates, random_state)
#              label_samples.append(X)
#
#          rate_samples = []
#          for i in range(num_rates):
#              X = []
#              for j, label in enumerate(self.labels):
#                  for row in label_samples[j][i]:
#                      X.append(row)
#              rate_samples.append(X)
#
#          datasets = []
#          for i in range(num_rates):
#              rate_samples[i] = shuffle_list(rate_samples[i],
#                                             random_state=random_state)
#              datasets.append(DualSentencesDataset(rate_samples[i]))
#
#          return datasets
#
#          #  samples = self.samples.to_numpy()
#          #  samples = shuffle_list(samples, random_state=random_state)
#          #
#          #  nos = [int(samples.shape[0] * r) for r in rates]
#          #  last = samples.shape[0] - np.sum(nos)
#          #  if last > 0:
#          #      nos += [last]
#          #  pos = np.add.accumulate(nos)
#          #
#          #  datasets = []
#          #  p0 = 0
#          #  for p1 in pos:
#          #      datasets.append(DualSentencesDataset(samples[p0:p1]))
#          #      p0 = p1
#          #
#          #  return datasets
#
#
#  #  class DualSentencesProcessor(DataProcessor):
#  #      def __init__(self, train_rate=0.9, train_fold=0):
#  #          self.ds_alldata = None
#  #          self.ds_test = None
#  #          self.train_rate = train_rate
#  #          self.fold = train_fold
#  #
#  #      def _load_train_dataset(self, data_dir):
#  #          self.ds_alldata = read_dataset(
#  #              os.path.join(data_dir, "train_dataset.csv"))
#  #          self.ds_train, self.ds_dev = self.ds_alldata.split(
#  #              rates=[self.train_rate], random_state=8864)
#  #
#  #          if self.fold > 0:
#  #              train_samples = self.ds_train.samples.to_numpy().tolist()
#  #              dev_samples = self.ds_dev.samples.to_numpy().tolist()
#  #              n = len(dev_samples)
#  #              p0 = n * (self.fold - 1)
#  #              p1 = p0 + n
#  #
#  #              #  tmp = train_samples[p0:p1]
#  #              #  train_samples[p0:p1] = dev_samples
#  #              #  dev_samples = tmp
#  #
#  #              train_samples[p0:p1], dev_samples = dev_samples, train_samples[
#  #                  p0:p1]
#  #
#  #              self.ds_train = DualSentencesDataset(train_samples)
#  #              self.ds_dev = DualSentencesDataset(dev_samples)
#  #
#  #      def _load_test_dataset(self, data_dir):
#  #          test_file = "test_dataset.csv"
#  #
#  #          test_data_file = os.path.join(data_dir, test_file)
#  #          print(f"test file: {test_file}")
#  #          test_data = pd.read_csv(test_data_file, sep='\t')
#  #          test_data.columns = ['sid', 'text_a', 'text_b', 'label', 'category']
#  #          test_data['label'] = ['0'] * test_data.shape[0]
#  #          test_data['category'] = ['0'] * test_data.shape[0]
#  #          self.ds_test = DualSentencesDataset(test_data)
#  #          #  self.ds_test = DualSentencesDataset()
#  #          #  self.ds_test.read_csv(test_data_file, default_label='0')
#  #
#  #      def get_example_from_tensor_dict(self, tensor_dict):
#  #          """See base class."""
#  #          return InputExample(tensor_dict['sid'].numpy(),
#  #                              tensor_dict['text_a'].numpy().decode('utf-8'),
#  #                              tensor_dict['text_b'].numpy().decode('utf-8'),
#  #                              str(tensor_dict['label'].numpy()))
#  #
#  #      def get_train_examples(self, data_dir):
#  #          if self.ds_alldata is None:
#  #              self._load_train_dataset(data_dir)
#  #          return self._create_examples(self.ds_train)
#  #
#  #      def get_dev_examples(self, data_dir):
#  #          if self.ds_alldata is None:
#  #              self._load_train_dataset(data_dir)
#  #          return self._create_examples(self.ds_dev)
#  #
#  #      def get_test_examples(self, data_dir):
#  #          if self.ds_test is None:
#  #              self._load_test_dataset(data_dir)
#  #          return self._create_examples(self.ds_test)
#  #
#  #      def get_labels(self):
#  #          return ["0", "1"]
#  #
#  #      def _create_examples(self, ds_examples):
#  #          examples = []
#  #          for (i, line) in enumerate(
#  #                  ds_examples.samples[['sid', 'text_a', 'text_b',
#  #                                       'label']].to_numpy()):
#  #              guid = line[0]
#  #              text_a = line[1]
#  #              text_b = line[2]
#  #              label = line[-1]
#  #              examples.append(
#  #                  InputExample(guid=guid,
#  #                               text_a=text_a,
#  #                               text_b=text_b,
#  #                               label=label))
#  #          return examples
